"""
Python-based grid handler, not to be confused with the SWIG-handler

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt.enzotools.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import exceptions
import pdb
import weakref

import numpy as np

from yt.funcs import *
from yt.utilities.definitions import x_dict, y_dict

from yt.data_objects.data_containers import YTFieldData
from yt.utilities.definitions import x_dict, y_dict
from .field_info_container import \
    NeedsGridType, \
    NeedsOriginalGrid, \
    NeedsDataField, \
    NeedsProperty, \
    NeedsParameter

class AMRGridPatch(object):
    _spatial = True
    _num_ghost_zones = 0
    _grids = None
    _id_offset = 1

    _type_name = 'grid'
    _skip_add = True
    _con_args = ('id', 'filename')
    OverlappingSiblings = None

    __slots__ = ['field_data', 'field_parameters', 'id', 'hierarchy', 'pf',
                 'ActiveDimensions', 'LeftEdge', 'RightEdge', 'Level',
                 'NumberOfParticles', 'Children', 'Parent',
                 'start_index', 'filename', '__weakref__', 'dds',
                 '_child_mask', '_child_indices', '_child_index_mask',
                 '_parent_id', '_children_ids']

    def __init__(self, id, filename=None, hierarchy=None):
        self.field_data = YTFieldData()
        self.field_parameters = {}
        self.id = id
        if hierarchy: self.hierarchy = weakref.proxy(hierarchy)
        self.pf = self.hierarchy.parameter_file  # weakref already
        self._child_mask = self._child_indices = self._child_index_mask = None
        self.start_index = None

    def get_global_startindex(self):
        """
        Return the integer starting index for each dimension at the current
        level.

        """
        if self.start_index is not None:
            return self.start_index
        if self.Parent == None:
            left = self.LeftEdge - self.pf.domain_left_edge
            start_index = left / self.dds
            return np.rint(start_index).astype('int64').ravel()

        pdx = self.Parent.dds
        start_index = (self.Parent.get_global_startindex()) + \
                       np.rint((self.LeftEdge - self.Parent.LeftEdge) / pdx)
        self.start_index = (start_index * self.pf.refine_by).astype('int64').ravel()
        return self.start_index

    def get_field_parameter(self, name, default=None):
        """
        This is typically only used by derived field functions, but it returns
        parameters used to generate fields.

        """
        if self.field_parameters.has_key(name):
            return self.field_parameters[name]
        else:
            return default

    def set_field_parameter(self, name, val):
        """
        Here we set up dictionaries that get passed up and down and ultimately
        to derived fields.

        """
        self.field_parameters[name] = val

    def has_field_parameter(self, name):
        """ Checks if a field parameter is set. """
        return self.field_parameters.has_key(name)

    def convert(self, datatype):
        """
        This will attempt to convert a given unit to cgs from code units. It
        either returns the multiplicative factor or throws a KeyError.

        """
        return self.pf[datatype]

    def __repr__(self):
        # We'll do this the slow way to be clear what's going on
        s = "%s (%s): " % (self.__class__.__name__, self.pf)
        s += ", ".join(["%s=%s" % (i, getattr(self,i))
                        for i in self._con_args])
        return s

    def _generate_field(self, field):
        if self.pf.field_info.has_key(field):
            # First we check the validator
            try:
                self.pf.field_info[field].check_available(self)
            except NeedsGridType, ngt_exception:
                # This is only going to be raised if n_gz > 0
                n_gz = ngt_exception.ghost_zones
                f_gz = ngt_exception.fields
                if f_gz is None:
                    f_gz = self.pf.field_info[field].get_dependencies(
                            pf = self.pf).requested
                gz_grid = self.retrieve_ghost_zones(n_gz, f_gz, smoothed=True)
                temp_array = self.pf.field_info[field](gz_grid)
                sl = [slice(n_gz, -n_gz)] * 3
                self[field] = temp_array[sl]
            else:
                self[field] = self.pf.field_info[field](self)
        else: # Can't find the field, try as it might
            raise exceptions.KeyError(field)

    def has_key(self, key):
        return (key in self.field_data)

    def __getitem__(self, key):
        """
        Returns a single field.  Will add if necessary.
        """
        if key not in self.field_data:
            self.get_data(key)
        return self.field_data[key]

    def __setitem__(self, key, val):
        """
        Sets a field to be some other value.
        """
        self.field_data[key] = val

    def __delitem__(self, key):
        """
        Deletes a field
        """
        del self.field_data[key]

    def keys(self):
        return self.field_data.keys()

    def get_data(self, field, convert = True):
        """
        Returns a field or set of fields for a key or set of keys
        """
        if not self.field_data.has_key(field):
            if field in self.hierarchy.field_list:
                conv_factor = 1.0
                if self.pf.field_info.has_key(field) and convert == True:
                    conv_factor = self.pf.field_info[field]._convert_function(self)
                if self.pf.field_info[field].particle_type and \
                   self.NumberOfParticles == 0:
                    # because this gets upcast to float
                    self[field] = np.array([],dtype='int64')
                    return self.field_data[field]
                try:
                    temp = self.hierarchy.io.pop(self, field)
                    self[field] = np.multiply(temp, conv_factor, temp)
                except self.hierarchy.io._read_exception, exc:
                    if field in self.pf.field_info:
                        if self.pf.field_info[field].not_in_all:
                            self[field] = np.zeros(self.ActiveDimensions, dtype='float64')
                        else:
                            raise
                    else: raise
            else:
                self._generate_field(field)
        return self.field_data[field]

    def _setup_dx(self):
        # So first we figure out what the index is.  We don't assume
        # that dx=dy=dz, at least here.  We probably do elsewhere.
        id = self.id - self._id_offset
        if self.Parent is not None:
            self.dds = self.Parent.dds / self.pf.refine_by
        else:
            LE, RE = self.hierarchy.grid_left_edge[id,:], \
                     self.hierarchy.grid_right_edge[id,:]
            self.dds = np.array((RE - LE) / self.ActiveDimensions)
        if self.pf.dimensionality < 2: self.dds[1] = self.pf.domain_right_edge[1] - self.pf.domain_left_edge[1]
        if self.pf.dimensionality < 3: self.dds[2] = self.pf.domain_right_edge[2] - self.pf.domain_left_edge[2]
        self.field_data['dx'], self.field_data['dy'], self.field_data['dz'] = self.dds

    @property
    def _corners(self):
        return np.array([ # Unroll!
            [self.LeftEdge[0],  self.LeftEdge[1],  self.LeftEdge[2]],
            [self.RightEdge[0], self.LeftEdge[1],  self.LeftEdge[2]],
            [self.RightEdge[0], self.RightEdge[1], self.LeftEdge[2]],
            [self.RightEdge[0], self.RightEdge[1], self.RightEdge[2]],
            [self.LeftEdge[0],  self.RightEdge[1], self.RightEdge[2]],
            [self.LeftEdge[0],  self.LeftEdge[1],  self.RightEdge[2]],
            [self.RightEdge[0], self.LeftEdge[1],  self.RightEdge[2]],
            [self.LeftEdge[0],  self.RightEdge[1], self.LeftEdge[2]],
            ], dtype='float64')

    def _generate_overlap_masks(self, axis, LE, RE):
        """
        Generate a mask that shows which cells overlap with arbitrary arrays
        *LE* and *RE*) of edges, typically grids, along *axis*.
        Use algorithm described at http://www.gamedev.net/reference/articles/article735.asp

        """
        x = x_dict[axis]
        y = y_dict[axis]
        cond = self.RightEdge[x] >= LE[:,x]
        cond = np.logical_and(cond, self.LeftEdge[x] <= RE[:,x])
        cond = np.logical_and(cond, self.RightEdge[y] >= LE[:,y])
        cond = np.logical_and(cond, self.LeftEdge[y] <= RE[:,y])
        return cond

    def is_in_grid(self, x, y, z) :
        """
        Generate a mask that shows which points in *x*, *y*, and *z*
        fall within this grid's boundaries.
        """
        xcond = np.logical_and(x >= self.LeftEdge[0],
                               x < self.RightEdge[0])
        ycond = np.logical_and(y >= self.LeftEdge[1],
                               y < self.RightEdge[1])
        zcond = np.logical_and(z >= self.LeftEdge[2],
                               z < self.RightEdge[2])

        cond = np.logical_and(xcond, ycond)
        cond = np.logical_and(zcond, cond)

        return cond
        
    def __repr__(self):
        return "AMRGridPatch_%04i" % (self.id)

    def __int__(self):
        return self.id

    def clear_data(self):
        """
        Clear out the following things: child_mask, child_indices, all fields,
        all field parameters.

        """
        self._del_child_mask()
        self._del_child_indices()
        self.field_data.clear()
        self._setup_dx()

    def check_child_masks(self):
        return self._child_mask, self._child_indices

    def _prepare_grid(self):
        """ Copies all the appropriate attributes from the hierarchy. """
        # This is definitely the slowest part of generating the hierarchy
        # Now we give it pointers to all of its attributes
        # Note that to keep in line with Enzo, we have broken PEP-8
        h = self.hierarchy # cache it
        my_ind = self.id - self._id_offset
        self.ActiveDimensions = h.grid_dimensions[my_ind]
        self.LeftEdge = h.grid_left_edge[my_ind]
        self.RightEdge = h.grid_right_edge[my_ind]
        h.grid_levels[my_ind, 0] = self.Level
        # This might be needed for streaming formats
        #self.Time = h.gridTimes[my_ind,0]
        self.NumberOfParticles = h.grid_particle_count[my_ind, 0]

    def __len__(self):
        return np.prod(self.ActiveDimensions)

    def find_max(self, field):
        """ Returns value, index of maximum value of *field* in this grid. """
        coord1d = (self[field] * self.child_mask).argmax()
        coord = np.unravel_index(coord1d, self[field].shape)
        val = self[field][coord]
        return val, coord

    def find_min(self, field):
        """ Returns value, index of minimum value of *field* in this grid. """
        coord1d = (self[field] * self.child_mask).argmin()
        coord = np.unravel_index(coord1d, self[field].shape)
        val = self[field][coord]
        return val, coord

    def get_position(self, index):
        """ Returns center position of an *index*. """
        pos = (index + 0.5) * self.dds + self.LeftEdge
        return pos

    def clear_all(self):
        """
        Clears all datafields from memory and calls
        :meth:`clear_derived_quantities`.

        """
        for key in self.keys():
            del self.field_data[key]
        del self.field_data
        if hasattr(self,"retVal"):
            del self.retVal
        self.field_data = YTFieldData()
        self.clear_derived_quantities()

    def clear_derived_quantities(self):
        """ Clears coordinates, child_indices, child_mask. """
        # Access the property raw-values here
        del self.child_mask
        del self.child_ind

    def _set_child_mask(self, newCM):
        if self._child_mask != None:
            mylog.warning("Overriding child_mask attribute!  This is probably unwise!")
        self._child_mask = newCM

    def _set_child_indices(self, newCI):
        if self._child_indices != None:
            mylog.warning("Overriding child_indices attribute!  This is probably unwise!")
        self._child_indices = newCI

    def _get_child_mask(self):
        if self._child_mask == None:
            self.__generate_child_mask()
        return self._child_mask

    def _get_child_indices(self):
        if self._child_indices == None:
            self.__generate_child_mask()
        return self._child_indices

    def _del_child_indices(self):
        try:
            del self._child_indices
        except AttributeError:
            pass
        self._child_indices = None

    def _del_child_mask(self):
        try:
            del self._child_mask
        except AttributeError:
            pass
        self._child_mask = None

    def _get_child_index_mask(self):
        if self._child_index_mask is None:
            self.__generate_child_index_mask()
        return self._child_index_mask

    def _del_child_index_mask(self):
        try:
            del self._child_index_mask
        except AttributeError:
            pass
        self._child_index_mask = None

    #@time_execution
    def __fill_child_mask(self, child, mask, tofill, dlevel = 1):
        rf = self.pf.refine_by
        if dlevel != 1:
            rf = rf**dlevel
        gi, cgi = self.get_global_startindex(), child.get_global_startindex()
        startIndex = np.maximum(0, cgi / rf - gi)
        endIndex = np.minimum((cgi + child.ActiveDimensions) / rf - gi,
                              self.ActiveDimensions)
        endIndex += (startIndex == endIndex)
        mask[startIndex[0]:endIndex[0],
             startIndex[1]:endIndex[1],
             startIndex[2]:endIndex[2]] = tofill

    def __generate_child_mask(self):
        """
        Generates self.child_mask, which is zero where child grids exist (and
        thus, where higher resolution data is available).

        """
        self._child_mask = np.ones(self.ActiveDimensions, 'int32')
        for child in self.Children:
            self.__fill_child_mask(child, self._child_mask, 0)
        if self.OverlappingSiblings is not None:
            for sibling in self.OverlappingSiblings:
                self.__fill_child_mask(sibling, self._child_mask, 0, 0)
        
        self._child_indices = (self._child_mask==0) # bool, possibly redundant

    def __generate_child_index_mask(self):
        """
        Generates self.child_index_mask, which is -1 where there is no child,
        and otherwise has the ID of the grid that resides there.

        """
        self._child_index_mask = np.zeros(self.ActiveDimensions, 'int32') - 1
        for child in self.Children:
            self.__fill_child_mask(child, self._child_index_mask,
                                   child.id)
        if self.OverlappingSiblings is not None:
            for sibling in self.OverlappingSiblings:
                self.__fill_child_mask(sibling, self._child_index_mask,
                                       sibling.id)

    def _get_coords(self):
        if self.__coords == None: self._generate_coords()
        return self.__coords

    def _set_coords(self, new_c):
        if self.__coords != None:
            mylog.warning("Overriding coords attribute!  This is probably unwise!")
        self.__coords = new_c

    def _del_coords(self):
        del self.__coords
        self.__coords = None

    def _generate_coords(self):
        """
        Creates self.coords, which is of dimensions (3, ActiveDimensions)

        """
        ind = np.indices(self.ActiveDimensions)
        left_shaped = np.reshape(self.LeftEdge, (3, 1, 1, 1))
        self['x'], self['y'], self['z'] = (ind + 0.5) * self.dds + left_shaped

    child_mask = property(fget=_get_child_mask, fdel=_del_child_mask)
    child_index_mask = property(fget=_get_child_index_mask, fdel=_del_child_index_mask)
    child_indices = property(fget=_get_child_indices, fdel = _del_child_indices)

    def retrieve_ghost_zones(self, n_zones, fields, all_levels=False,
                             smoothed=False):
        # We will attempt this by creating a datacube that is exactly bigger
        # than the grid by nZones*dx in each direction
        nl = self.get_global_startindex() - n_zones
        nr = nl + self.ActiveDimensions + 2 * n_zones
        new_left_edge = nl * self.dds + self.pf.domain_left_edge
        new_right_edge = nr * self.dds + self.pf.domain_left_edge

        # Something different needs to be done for the root grid, though
        level = self.Level
        if all_levels:
            level = self.hierarchy.max_level + 1
        args = (level, new_left_edge, new_right_edge)
        kwargs = {'dims': self.ActiveDimensions + 2*n_zones,
                  'num_ghost_zones':n_zones,
                  'use_pbar':False, 'fields':fields}
        # This should update the arguments to set the field parameters to be
        # those of this grid.
        kwargs.update(self.field_parameters)
        if smoothed:
            cube = self.hierarchy.smoothed_covering_grid(
                level, new_left_edge, **kwargs)
        else:
            cube = self.hierarchy.covering_grid(level, new_left_edge, **kwargs)

        return cube

    def get_vertex_centered_data(self, field, smoothed=True, no_ghost=False):
        new_field = np.zeros(self.ActiveDimensions + 1, dtype='float64')

        if no_ghost:
            of = self[field]
            new_field[:-1,:-1,:-1] += of
            new_field[:-1,:-1,1:] += of
            new_field[:-1,1:,:-1] += of
            new_field[:-1,1:,1:] += of
            new_field[1:,:-1,:-1] += of
            new_field[1:,:-1,1:] += of
            new_field[1:,1:,:-1] += of
            new_field[1:,1:,1:] += of
            np.multiply(new_field, 0.125, new_field)
            if self.pf.field_info[field].take_log:
                new_field = np.log10(new_field)

            new_field[:,:, -1] = 2.0*new_field[:,:,-2] - new_field[:,:,-3]
            new_field[:,:, 0]  = 2.0*new_field[:,:,1] - new_field[:,:,2]
            new_field[:,-1, :] = 2.0*new_field[:,-2,:] - new_field[:,-3,:]
            new_field[:,0, :]  = 2.0*new_field[:,1,:] - new_field[:,2,:]
            new_field[-1,:,:] = 2.0*new_field[-2,:,:] - new_field[-3,:,:]
            new_field[0,:,:]  = 2.0*new_field[1,:,:] - new_field[2,:,:]

            if self.pf.field_info[field].take_log:
                np.power(10.0, new_field, new_field)
        else:
            cg = self.retrieve_ghost_zones(1, field, smoothed=smoothed)
            np.add(new_field, cg[field][1: ,1: ,1: ], new_field)
            np.add(new_field, cg[field][:-1,1: ,1: ], new_field)
            np.add(new_field, cg[field][1: ,:-1,1: ], new_field)
            np.add(new_field, cg[field][1: ,1: ,:-1], new_field)
            np.add(new_field, cg[field][:-1,1: ,:-1], new_field)
            np.add(new_field, cg[field][1: ,:-1,:-1], new_field)
            np.add(new_field, cg[field][:-1,:-1,1: ], new_field)
            np.add(new_field, cg[field][:-1,:-1,:-1], new_field)
            np.multiply(new_field, 0.125, new_field)

        return new_field



from yt.mods import *
import matplotlib
import pylab
from output_tests import SingleOutputTest, YTStaticOutputTest, create_test
from yt.analysis_modules.halo_finding.api import *
import hashlib
import numpy as np

# Tests the number of halos returned by the HOP halo finder on a dataset
class TestHaloCountHOP(YTStaticOutputTest):
    threshold = 80.0

    def run(self):
        # Find the halos using vanilla HOP.
        halos = HaloFinder(self.pf, threshold=self.threshold, dm_only=False)
        # We only care about the number of halos.
        self.result = len(halos)
                    
    def compare(self, old_result):
        # The new value should be identical to the old one.
        self.compare_value_delta(self.result, old_result, 0)

    def plot(self):
        return []

# Tests the number of halos returned by the FOF halo finder on a dataset
class TestHaloCountFOF(YTStaticOutputTest):
    link = 0.2
    padding = 0.02

    def run(self):
        # Find the halos using FOF.
        halos = FOFHaloFinder(self.pf, link=self.link, dm_only=False, 
                               padding=self.padding)
        # We only care about the number of halos.
        self.result = len(halos)
                    
    def compare(self, old_result):
        # The new value should be identical to the old one.
        self.compare_value_delta(self.result, old_result, 0)

    def plot(self):
        return []

# Tests the number of halos returned by the Parallel HOP halo finder on a 
# dataset
class TestHaloCountPHOP(YTStaticOutputTest):
    threshold = 80.0

    def run(self):
        # Find the halos using parallel HOP.
        halos = parallelHF(self.pf, threshold=self.threshold, dm_only=False)
        # We only care about the number of halos.
        self.result = len(halos)
                    
    def compare(self, old_result):
        # The new value should be identical to the old one.
        self.compare_value_delta(self.result, old_result, 0)

    def plot(self):
        return []

class TestHaloComposition(YTStaticOutputTest):
    threshold=80.0
    
    def run(self):
        # Find the halos using vanilla HOP.
        halos = HaloFinder(self.pf, threshold=self.threshold, dm_only=False)
        # The result is a list of the particle IDs, stored
        # as sets for easy comparison.
        IDs = []
        for halo in halos:
            IDs.append(set(halo["particle_index"]))
        self.result = IDs
    
    def compare(self, old_result):
        # All the sets should be identical.
        pairs = zip(self.result, old_result)
        for pair in pairs:
            if len(pair[0] - pair[1]) != 0:
                return False
        return True
    
# Tests the content of the halos returned by the HOP halo finder on a dataset 
# by comparing the hash of the arrays of all the particles contained in each
# halo.  Evidently breaks on parallel runtime.  DO NOT USE.
class TestHaloCompositionHashHOP(YTStaticOutputTest):
    threshold=80.0
    
    def run(self):
        # Find the halos using vanilla HOP.
        halos = HaloFinder(self.pf, threshold=self.threshold, dm_only=False)
        # The result is a flattened array of the arrays of the particle IDs for
        # each halo
        IDs = []
        for halo in halos:
            IDs.append(halo["particle_index"])
        IDs = np.concatenate(IDs)
        self.result = IDs
    
    def compare(self, old_result):
        # All the lists of arrays should be identical.  To check this
        # faster, we take the 256-bit hash of these lists and compare them
        result_hash = hashlib.sha256(self.result.tostring()).hexdigest()
        old_result_hash = hashlib.sha256(old_result.tostring()).hexdigest()
        if result_hash == old_result_hash:
            return True
        else:
            return False

# Tests the content of the halos returned by the FOF halo finder on a dataset 
# by comparing the hash of the arrays of all the particles contained in each
# halo.  Evidently breaks on parallel runtime.  DO NOT USE.
class TestHaloCompositionHashFOF(YTStaticOutputTest):
    link = 0.2
    padding = 0.02
    
    def run(self):
        # Find the halos using vanilla FOF.
        halos = FOFHaloFinder(self.pf, link=self.link, dm_only=False, 
                               padding=self.padding)
        # The result is a flattened array of the arrays of the particle IDs for
        # each halo
        IDs = []
        for halo in halos:
            IDs.append(halo["particle_index"])
        IDs = np.concatenate(IDs)
        self.result = IDs
    
    def compare(self, old_result):
        # All the lists of arrays should be identical.  To check this
        # faster, we take the 256-bit hash of these lists and compare them
        result_hash = hashlib.sha256(self.result.tostring()).hexdigest()
        old_result_hash = hashlib.sha256(old_result.tostring()).hexdigest()
        if result_hash == old_result_hash:
            return True
        else:
            return False

# Tests the content of the halos returned by the Parallel HOP halo finder on a 
# dataset by comparing the hash of the arrays of all the particles contained 
# in each halo.  Evidently breaks on parallel runtime.  DO NOT USE.
class TestHaloCompositionHashPHOP(YTStaticOutputTest):
    threshold=80.0
    
    def run(self):
        # Find the halos using parallel HOP.
        halos = parallelHF(self.pf, threshold=self.threshold, dm_only=False)
        # The result is a flattened array of the arrays of the particle IDs for
        # each halo
        IDs = []
        for halo in halos:
            IDs.append(halo["particle_index"])
        IDs = np.concatenate(IDs)
        self.result = IDs
    
    def compare(self, old_result):
        # All the lists of arrays should be identical.  To check this
        # faster, we take the 256-bit hash of these lists and compare them
        result_hash = hashlib.sha256(self.result.tostring()).hexdigest()
        old_result_hash = hashlib.sha256(old_result.tostring()).hexdigest()
        if result_hash == old_result_hash:
            return True
        else:
            return False



import matplotlib
from yt.mods import *
import pylab
from output_tests import SingleOutputTest, YTStaticOutputTest, create_test

class TestParticleUniqueIDs(YTStaticOutputTest):

    def run(self):
        # Test to make sure that all the particles have unique IDs.
        all = self.pf.h.all_data()
        IDs = all["particle_index"]
        # Make sure the order is the same every time.
        IDs = IDs[IDs.argsort()]
        self.result = IDs
                    
    def compare(self, old_result):
        # Two things: there should be no repeats in either the new or
        # the old, and the two sets should be the same.
        if len(old_result) != len(set(old_result)): return False
        if len(self.result) != len(set(self.result)): return False
        if (self.result != old_result).all(): return False
        return True

    def plot(self):
        return []

create_test(TestParticleUniqueIDs, "particle_unique_ids_test")

class TestParticleExtrema(YTStaticOutputTest):

    def run(self):
        # Tests to make sure there are no particle positions aren't changing
        # drastically. This is very unlikely to be a problem.
        all = self.pf.h.all_data()
        min = np.empty(3,dtype='float64')
        max = min.copy()
        dims = ["particle_position_x","particle_position_y",
            "particle_position_z"]
        for i in xrange(3):
            min[i] = np.min(all[dims[i]])
            max[i] = np.max(all[dims[i]])
        self.result = (min,max)
    
    def compare(self, old_result):
        min,max = self.result
        old_min, old_max = old_result
        # The extrema should be very similar.
        self.compare_array_delta(min, old_min, 1e-7)
        self.compare_array_delta(max, old_max, 1e-7)
        # Also, the min/max shouldn't be outside the boundaries.
        if (min < self.pf.domain_left_edge).any(): return False
        if (max > self.pf.domain_right_edge).any(): return False
        return True
    
    def plot(self):
        return []

create_test(TestParticleExtrema, "particle_extrema_test")




"""
Definitions specific to ART

Author: Christopher E. Moody <cemoody@ucsc.edu>
Affiliation: UC Santa Cruz
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Christopher E. Moody.  All Rights
  Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""

fluid_fields= [ 
    'Density',
    'TotalEnergy',
    'XMomentumDensity',
    'YMomentumDensity',
    'ZMomentumDensity',
    'Pressure',
    'Gamma',
    'GasEnergy',
    'MetalDensitySNII',
    'MetalDensitySNIa',
    'PotentialNew',
    'PotentialOld'
]

particle_fields= [
    'particle_age',
    'particle_index',
    'particle_mass',
    'particle_mass_initial',
    'particle_creation_time',
    'particle_metallicity1',
    'particle_metallicity2',
    'particle_metallicity',
    'particle_position_x',
    'particle_position_y',
    'particle_position_z',
    'particle_velocity_x',
    'particle_velocity_y',
    'particle_velocity_z',
    'particle_type'
]

particle_star_fields = [
    'particle_age',
    'particle_mass',
    'particle_mass_initial',
    'particle_creation_time',
    'particle_metallicity1',
    'particle_metallicity2',
    'particle_metallicity',
]

filename_pattern = {				
	'amr':'10MpcBox_csf512_%s.d',
	'particle_header':'PMcrd%s.DAT',
	'particle_data':'PMcrs0%s.DAT',
	'particle_stars':'stars_%s.dat'
}

amr_header_struct = [
    ('>i','pad byte'),
    ('>256s','jname'),
    ('>i','pad byte'),
    ('>i','pad byte'),
    ('>i','istep'),
    ('>d','t'),
    ('>d','dt'),
    ('>f','aexpn'),
    ('>f','ainit'),
    ('>i','pad byte'),
    ('>i','pad byte'),
    ('>f','boxh'),
    ('>f','Om0'),
    ('>f','Oml0'),
    ('>f','Omb0'),
    ('>f','hubble'),
    ('>i','pad byte'),
    ('>i','pad byte'),
    ('>i','nextras'),
    ('>i','pad byte'),
    ('>i','pad byte'),
    ('>f','extra1'),
    ('>f','extra2'),
    ('>i','pad byte'),
    ('>i','pad byte'),
    ('>256s','lextra'),
    ('>256s','lextra'),
    ('>i','pad byte'),
    ('>i', 'pad byte'),
    ('>i', 'min_level'),
    ('>i', 'max_level'),
    ('>i', 'pad byte'),
]

particle_header_struct =[
    ('>i','pad'),
    ('45s','header'), 
    ('>f','aexpn'),
    ('>f','aexp0'),
    ('>f','amplt'),
    ('>f','astep'),
    ('>i','istep'),
    ('>f','partw'),
    ('>f','tintg'),
    ('>f','Ekin'),
    ('>f','Ekin1'),
    ('>f','Ekin2'),
    ('>f','au0'),
    ('>f','aeu0'),
    ('>i','Nrow'),
    ('>i','Ngridc'),
    ('>i','Nspecies'),
    ('>i','Nseed'),
    ('>f','Om0'),
    ('>f','Oml0'),
    ('>f','hubble'),
    ('>f','Wp5'),
    ('>f','Ocurv'),
    ('>f','Omb0'),
    ('>%ds'%(396),'extras'),
    ('>f','unknown'),
    ('>i','pad')
]

constants = {
    "Y_p":0.245,
    "gamma":5./3.,
    "T_CMB0":2.726,
    "T_min":300.,
    "ng":128,
    "wmu":4.0/(8.0-5.0*0.245)
}

seek_extras = 137



"""
Nyx data-file handling functions (basically a boxlib reader)

Author: Casey W. Stark <caseywstark@gmail.com>
Affiliation: UC Berkeley
Author: Matthew Turk <matthewturk@gmail.com>
Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Casey W. Stark, Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import numpy as np
from yt.utilities.lib import read_castro_particles, read_and_seek
from yt.utilities.io_handler import BaseIOHandler

from definitions import fab_header_pattern, nyx_particle_field_names, \
                        yt_to_nyx_fields_dict

class IOHandlerNative(BaseIOHandler):
    """ File handler that can somehow read the native boxlib format. """

    _data_style = "nyx_native"

    def modify(self, field):
        return field.swapaxes(0, 2)

    def _read_particle_field(self, grid, field):
        offset = grid._particle_offset
        filen = os.path.expanduser(grid.particle_filename)
        off = grid._particle_offset
        tr = np.zeros(grid.NumberOfParticles, dtype='float64')
        read_castro_particles(filen, off,
                              nyx_particle_field_names.index(field),
                              len(nyx_particle_field_names), tr)
        return tr

    def _read_data(self, grid, field):
        """ reads packed multiFABs output by BoxLib in "NATIVE" format. """
        if field in nyx_particle_field_names:
            return self._read_particle_field(grid, field)
        filen = os.path.expanduser(grid.filename[field])
        offset1 = grid._offset[field]
        # one field has nElements * bytesPerReal bytes and is located
        # nElements * bytesPerReal * field_index from the offset location
        bytesPerReal = grid.hierarchy._bytesPerReal

        fieldname = yt_to_nyx_fields_dict.get(field, field)
        field_index = grid.field_indexes[fieldname]
        nElements = grid.ActiveDimensions.prod()
        offset2 = int(nElements*bytesPerReal*field_index)

        dtype = grid.hierarchy._dtype
        field = np.empty(nElements, dtype=grid.hierarchy._dtype)
        read_and_seek(filen, offset1, offset2, field, nElements * bytesPerReal)
        field = field.reshape(grid.ActiveDimensions, order='F')

        # @todo: we can/should also check against the max and min in the header
        # file

        return field




"""
The basic field info container resides here.  These classes, code specific and
universal, are the means by which we access fields across YT, both derived and
native.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import types
import inspect
import copy
import itertools

import numpy as np

from yt.funcs import *

class FieldInfoContainer(dict): # Resistance has utility
    """
    This is a generic field container.  It contains a list of potential derived
    fields, all of which know how to act on a data object and return a value.
    This object handles converting units as well as validating the availability
    of a given field.

    """
    fallback = None

    def add_field(self, name, function=None, **kwargs):
        """
        Add a new field, along with supplemental metadata, to the list of
        available fields.  This respects a number of arguments, all of which
        are passed on to the constructor for
        :class:`~yt.data_objects.api.DerivedField`.

        """
        if function == None:
            def create_function(function):
                self[name] = DerivedField(name, function, **kwargs)
                return function
            return create_function
        self[name] = DerivedField(name, function, **kwargs)
        
    def add_grad(self, field, **kwargs):
        """
        Creates the partial derivative of a given field. This function will
        autogenerate the names of the gradient fields.

        """
        sl = slice(2,None,None)
        sr = slice(None,-2,None)
        
        def _gradx(f, data):
            grad = data[field][sl,1:-1,1:-1] - data[field][sr,1:-1,1:-1]
            grad /= 2.0*data["dx"].flat[0]
            g = np.zeros(data[field].shape, dtype='float64')
            g[1:-1,1:-1,1:-1] = grad
            return g
            
        def _grady(f, data):
            grad = data[field][1:-1,sl,1:-1] - data[field][1:-1,sr,1:-1]
            grad /= 2.0*data["dy"].flat[0]
            g = np.zeros(data[field].shape, dtype='float64')
            g[1:-1,1:-1,1:-1] = grad
            return g
            
        def _gradz(f, data):
            grad = data[field][1:-1,1:-1,sl] - data[field][1:-1,1:-1,sr]
            grad /= 2.0*data["dz"].flat[0]
            g = np.zeros(data[field].shape, dtype='float64')
            g[1:-1,1:-1,1:-1] = grad
            return g
        
        d_kwargs = kwargs.copy()
        if "display_name" in kwargs: del d_kwargs["display_name"]
        
        for ax in "xyz":
            if "display_name" in kwargs:
                disp_name = r"%s\_%s" % (kwargs["display_name"], ax)
            else:
                disp_name = r"\partial %s/\partial %s" % (field, ax)
            name = "Grad_%s_%s" % (field, ax)
            self[name] = DerivedField(name, function=eval('_grad%s' % ax),
                         take_log=False, validators=[ValidateSpatial(1,[field])],
                         display_name = disp_name, **d_kwargs)
        
        def _grad(f, data) :
            a = np.power(data["Grad_%s_x" % field],2)
            b = np.power(data["Grad_%s_y" % field],2)
            c = np.power(data["Grad_%s_z" % field],2)
            norm = np.sqrt(a+b+c)
            return norm

        if "display_name" in kwargs:
            disp_name = kwargs["display_name"]
        else:
            disp_name = r"\Vert\nabla %s\Vert" % (field)   
        name = "Grad_%s" % field           
        self[name] = DerivedField(name, function=_grad, take_log=False,
                                  display_name = disp_name, **d_kwargs)
        mylog.info("Added new fields: Grad_%s_x, Grad_%s_y, Grad_%s_z, Grad_%s" \
                   % (field, field, field, field))

    def has_key(self, key):
        # This gets used a lot
        if key in self: return True
        if self.fallback is None: return False
        return self.fallback.has_key(key)

    def __missing__(self, key):
        if self.fallback is None:
            raise KeyError("No field named %s" % key)
        return self.fallback[key]

    @classmethod
    def create_with_fallback(cls, fallback):
        obj = cls()
        obj.fallback = fallback
        return obj

    def __contains__(self, key):
        if dict.__contains__(self, key): return True
        if self.fallback is None: return False
        return self.fallback.has_key(key)

    def __iter__(self):
        for f in dict.__iter__(self): yield f
        if self.fallback:
            for f in self.fallback: yield f

def TranslationFunc(field_name):
    def _TranslationFunc(field, data):
        return data[field_name]
    return _TranslationFunc

def NullFunc(field, data):
    return

FieldInfo = FieldInfoContainer()
add_field = FieldInfo.add_field
add_grad = FieldInfo.add_grad

def derived_field(**kwargs):
    def inner_decorator(function):
        if 'name' not in kwargs:
            kwargs['name'] = function.func_name
        kwargs['function'] = function
        add_field(**kwargs)
        return function
    return inner_decorator

class ValidationException(Exception):
    pass

class NeedsGridType(ValidationException):
    def __init__(self, ghost_zones = 0, fields=None):
        self.ghost_zones = ghost_zones
        self.fields = fields
    def __str__(self):
        return "(%s, %s)" % (self.ghost_zones, self.fields)

class NeedsOriginalGrid(NeedsGridType):
    def __init__(self):
        self.ghost_zones = 0

class NeedsDataField(ValidationException):
    def __init__(self, missing_fields):
        self.missing_fields = missing_fields
    def __str__(self):
        return "(%s)" % (self.missing_fields)

class NeedsProperty(ValidationException):
    def __init__(self, missing_properties):
        self.missing_properties = missing_properties
    def __str__(self):
        return "(%s)" % (self.missing_properties)

class NeedsParameter(ValidationException):
    def __init__(self, missing_parameters):
        self.missing_parameters = missing_parameters
    def __str__(self):
        return "(%s)" % (self.missing_parameters)

class FieldDetector(defaultdict):
    Level = 1
    NumberOfParticles = 1
    _read_exception = None
    _id_offset = 0

    def __init__(self, nd = 16, pf = None, flat = False):
        self.nd = nd
        self.flat = flat
        self._spatial = not flat
        self.ActiveDimensions = [nd,nd,nd]
        self.LeftEdge = [0.0, 0.0, 0.0]
        self.RightEdge = [1.0, 1.0, 1.0]
        self.dds = np.ones(3, "float64")
        self['dx'] = self['dy'] = self['dz'] = np.array([1.0])
        class fake_parameter_file(defaultdict):
            pass

        if pf is None:
            # required attrs
            pf = fake_parameter_file(lambda: 1)
            pf.current_redshift = pf.omega_lambda = pf.omega_matter = \
                pf.cosmological_simulation = 0.0
            pf.hubble_constant = 0.7
            pf.domain_left_edge = np.zeros(3, 'float64')
            pf.domain_right_edge = np.ones(3, 'float64')
            pf.dimensionality = 3
            pf.periodicity = (True, True, True)
        self.pf = pf

        class fake_hierarchy(object):
            class fake_io(object):
                def _read_data_set(io_self, data, field):
                    return self._read_data(field)
                _read_exception = RuntimeError
            io = fake_io()
            def get_smallest_dx(self):
                return 1.0

        self.hierarchy = fake_hierarchy()
        self.requested = []
        self.requested_parameters = []
        if not self.flat:
            defaultdict.__init__(self,
                lambda: np.ones((nd, nd, nd), dtype='float64')
                + 1e-4*np.random.random((nd, nd, nd)))
        else:
            defaultdict.__init__(self, 
                lambda: np.ones((nd * nd * nd), dtype='float64')
                + 1e-4*np.random.random((nd * nd * nd)))

    def __missing__(self, item):
        FI = getattr(self.pf, "field_info", FieldInfo)
        if FI.has_key(item) and FI[item]._function.func_name != 'NullFunc':
            try:
                vv = FI[item](self)
            except NeedsGridType as exc:
                ngz = exc.ghost_zones
                nfd = FieldDetector(self.nd + ngz * 2)
                nfd._num_ghost_zones = ngz
                vv = FI[item](nfd)
                if ngz > 0: vv = vv[ngz:-ngz, ngz:-ngz, ngz:-ngz]
                for i in nfd.requested:
                    if i not in self.requested: self.requested.append(i)
                for i in nfd.requested_parameters:
                    if i not in self.requested_parameters:
                        self.requested_parameters.append(i)
            if vv is not None:
                if not self.flat: self[item] = vv
                else: self[item] = vv.ravel()
                return self[item]
        self.requested.append(item)
        if item not in self:
            self[item] = self._read_data(item)
        return self[item]

    def _read_data(self, field_name):
        self.requested.append(field_name)
        FI = getattr(self.pf, "field_info", FieldInfo)
        if field_name in FI and FI[field_name].particle_type:
            self.requested.append(field_name)
            return np.ones(self.NumberOfParticles)
        return defaultdict.__missing__(self, field_name)

    def get_field_parameter(self, param, default = None):
        self.requested_parameters.append(param)
        if param in ['bulk_velocity', 'center', 'normal']:
            return np.random.random(3) * 1e-2
        else:
            return 0.0
    _num_ghost_zones = 0
    id = 1
    def has_field_parameter(self, param): return True
    def convert(self, item): return 1

class DerivedField(object):
    """
    This is the base class used to describe a cell-by-cell derived field.

    Parameters
    ----------

    name : str
       is the name of the field.
    function : callable
       A function handle that defines the field.  Should accept
       arguments (field, data)
    convert_function : callable
       A function that converts to CGS, **only if necessary**
    units : str
       A mathtext-formatted string that describes the field
    projected_units : str
       If we display a projection, what should the units be?
    take_log : bool
       Describes whether the field should be logged
    validators : list
       A list of :class:`FieldValidator` objects
    particle_type : bool
       Is this a particle (1D) field?
    vector_field : bool
       Describes the dimensionality of the field.  Currently unused.
    display_field : bool
       Governs its appearance in the dropdowns in Reason
    not_in_all : bool
       Used for baryon fields from the data that are not in all the grids
    display_name : str
       A name used in the plots
    projection_conversion : unit
       which unit should we multiply by in a projection?
    """
    def __init__(self, name, function,
                 convert_function = None,
                 particle_convert_function = None,
                 units = "", projected_units = "",
                 take_log = True, validators = None,
                 particle_type = False, vector_field=False,
                 display_field = True, not_in_all=False,
                 display_name = None, projection_conversion = "cm"):
        self.name = name
        self._function = function
        if validators:
            self.validators = ensure_list(validators)
        else:
            self.validators = []
        self.take_log = take_log
        self._units = units
        self._projected_units = projected_units
        if not convert_function:
            convert_function = lambda a: 1.0
        self._convert_function = convert_function
        self._particle_convert_function = particle_convert_function
        self.particle_type = particle_type
        self.vector_field = vector_field
        self.projection_conversion = projection_conversion
        self.display_field = display_field
        self.display_name = display_name
        self.not_in_all = not_in_all

    def check_available(self, data):
        """
        This raises an exception of the appropriate type if the set of
        validation mechanisms are not met, and otherwise returns True.
        """
        for validator in self.validators:
            validator(data)
        # If we don't get an exception, we're good to go
        return True

    def get_dependencies(self, *args, **kwargs):
        """
        This returns a list of names of fields that this field depends on.
        """
        e = FieldDetector(*args, **kwargs)
        if self._function.func_name == '<lambda>':
            e.requested.append(self.name)
        else:
            e[self.name]
        return e

    def get_units(self):
        """ Return a string describing the units. """
        return self._units

    def get_projected_units(self):
        """
        Return a string describing the units if the field has been projected.
        """
        return self._projected_units

    def __call__(self, data):
        """ Return the value of the field in a given *data* object. """
        ii = self.check_available(data)
        original_fields = data.keys() # Copy
        dd = self._function(self, data)
        dd *= self._convert_function(data)
        for field_name in data.keys():
            if field_name not in original_fields:
                del data[field_name]
        return dd

    def get_source(self):
        """
        Return a string containing the source of the function (if possible.)
        """
        return inspect.getsource(self._function)

    def get_label(self, projected=False):
        """
        Return a data label for the given field, inluding units.
        """
        name = self.name
        if self.display_name is not None: name = self.display_name
        data_label = r"$\rm{%s}" % name
        if projected: units = self.get_projected_units()
        else: units = self.get_units()
        if units != "": data_label += r"\/\/ (%s)" % (units)
        data_label += r"$"
        return data_label

    def particle_convert(self, data):
        if self._particle_convert_function is not None:
            return self._particle_convert_function(data)
        return None

class FieldValidator(object):
    pass

class ValidateParameter(FieldValidator):
    def __init__(self, parameters):
        """
        This validator ensures that the parameter file has a given parameter.
        """
        FieldValidator.__init__(self)
        self.parameters = ensure_list(parameters)
    def __call__(self, data):
        doesnt_have = []
        for p in self.parameters:
            if not data.has_field_parameter(p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsParameter(doesnt_have)
        return True

class ValidateDataField(FieldValidator):
    def __init__(self, field):
        """
        This validator ensures that the output file has a given data field stored
        in it.
        """
        FieldValidator.__init__(self)
        self.fields = ensure_list(field)
    def __call__(self, data):
        doesnt_have = []
        if isinstance(data, FieldDetector): return True
        for f in self.fields:
            if f not in data.hierarchy.field_list:
                doesnt_have.append(f)
        if len(doesnt_have) > 0:
            raise NeedsDataField(doesnt_have)
        return True

class ValidateProperty(FieldValidator):
    def __init__(self, prop):
        """
        This validator ensures that the data object has a given python attribute.
        """
        FieldValidator.__init__(self)
        self.prop = ensure_list(prop)
    def __call__(self, data):
        doesnt_have = []
        for p in self.prop:
            if not hasattr(data,p):
                doesnt_have.append(p)
        if len(doesnt_have) > 0:
            raise NeedsProperty(doesnt_have)
        return True

class ValidateSpatial(FieldValidator):
    def __init__(self, ghost_zones = 0, fields=None):
        """
        This validator ensures that the data handed to the field is of spatial
        nature -- that is to say, 3-D.
        """
        FieldValidator.__init__(self)
        self.ghost_zones = ghost_zones
        self.fields = fields
    def __call__(self, data):
        # When we say spatial information, we really mean
        # that it has a three-dimensional data structure
        #if isinstance(data, FieldDetector): return True
        if not data._spatial:
            raise NeedsGridType(self.ghost_zones,self.fields)
        if self.ghost_zones <= data._num_ghost_zones:
            return True
        raise NeedsGridType(self.ghost_zones,self.fields)

class ValidateGridType(FieldValidator):
    def __init__(self):
        """
        This validator ensures that the data handed to the field is an actual
        grid patch, not a covering grid of any kind.
        """
        FieldValidator.__init__(self)
    def __call__(self, data):
        # We need to make sure that it's an actual AMR grid
        if isinstance(data, FieldDetector): return True
        if getattr(data, "_type_name", None) == 'grid': return True
        raise NeedsOriginalGrid()



"""
Automagical cartesian domain decomposition.

Author: Kacper Kowalik <xarthisius.kk@gmail.com>
Affiliation: CA UMK
Author: Artur Gawryszczak <gawrysz@gmail.com>
Affiliation: PCSS
Homepage: http://yt-project.org/
License:
  Copyright (C) 2012 Kacper Kowalik. All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

SIEVE_PRIMES = \
    lambda l: l and l[:1] + SIEVE_PRIMES([n for n in l if n % l[0]])


def decompose_to_primes(max_prime):
    """ Decompose number into the primes """
    for prime in SIEVE_PRIMES(range(2, max_prime)):
        if prime * prime > max_prime:
            break
        while max_prime % prime == 0:
            yield prime
            max_prime /= prime
    if max_prime > 1:
        yield max_prime


def decompose_array(arr, psize, bbox):
    """ Calculate list of product(psize) subarrays of arr, along with their
        left and right edges
    """
    grid_left_edges = np.empty([np.product(psize), 3], dtype=np.float64)
    grid_right_edges = np.empty([np.product(psize), 3], dtype=np.float64)
    n_d = arr.shape
    d_s = (bbox[:, 1] - bbox[:, 0]) / n_d
    dist = np.mgrid[bbox[0, 0]:bbox[0, 1]:d_s[0],
                    bbox[1, 0]:bbox[1, 1]:d_s[1],
                    bbox[2, 0]:bbox[2, 1]:d_s[2]]
    for i in range(3):
        xyz = split_array(dist[i], psize)
        for j in range(np.product(psize)):
            grid_left_edges[j, i] = xyz[j][0, 0, 0]
            grid_right_edges[j, i] = xyz[j][-1, -1, -1] + d_s[i]
        del xyz
    del dist
    patches = split_array(arr, psize)
    return grid_left_edges, grid_right_edges, patches


def evaluate_domain_decomposition(n_d, pieces, ldom):
    """ Evaluate longest to shortest edge ratio
        BEWARE: lot's of magic here """
    eff_dim = (n_d > 1).sum()
    exp = float(eff_dim - 1) / float(eff_dim)
    ideal_bsize = eff_dim * pieces ** (1.0 / eff_dim) * np.product(n_d) ** exp
    mask = np.where(n_d > 1)
    nd_arr = np.array(n_d, dtype=np.float64)[mask]
    bsize = int(np.sum(ldom[mask] / nd_arr * np.product(nd_arr)))
    load_balance = float(np.product(n_d)) / \
        (float(pieces) * np.product((n_d - 1) / ldom + 1))

    # 0.25 is magic number
    quality = load_balance / (1 + 0.25 * (bsize / ideal_bsize - 1.0))
    # \todo add a factor that estimates lower cost when x-direction is
    # not chopped too much
    # \deprecated estimate these magic numbers
    quality *= (1. - (0.001 * ldom[0] + 0.0001 * ldom[1]) / pieces)
    if np.any(ldom > n_d):
        quality = 0

    return quality


def factorize_number(pieces):
    """ Return array consiting of prime, its power and number of different
        decompositions in three dimensions for this prime
    """
    factors = [factor for factor in decompose_to_primes(pieces)]
    temp = np.bincount(factors)
    return np.array(
        [(prime, temp[prime], (temp[prime] + 1) * (temp[prime] + 2) / 2)
         for prime in np.unique(factors)]
    )


def get_psize(n_d, pieces):
    """ Calculate the best division of array into px*py*pz subarrays.
        The goal is to minimize the ratio of longest to shortest edge
        to minimize the amount of inter-process communication.
    """
    fac = factorize_number(pieces)
    nfactors = len(fac[:, 2])
    best = 0.0
    p_size = np.ones(3, dtype=np.int)
    if pieces == 1:
        return p_size

    while np.all(fac[:, 2] > 0):
        ldom = np.ones(3, dtype=np.int)
        for nfac in range(nfactors):
            i = int(np.sqrt(0.25 + 2 * (fac[nfac, 2] - 1)) - 0.5)
            k = fac[nfac, 2] - int(1 + i * (i + 1) / 2)
            i = fac[nfac, 1] - i
            j = fac[nfac, 1] - (i + k)
            ldom *= fac[nfac, 0] ** np.array([i, j, k])

        quality = evaluate_domain_decomposition(n_d, pieces, ldom)
        if quality > best:
            best = quality
            p_size = ldom
        # search for next unique combination
        for j in range(nfactors):
            if fac[j, 2] > 1:
                fac[j, 2] -= 1
                break
            else:
                if (j < nfactors - 1):
                    fac[j, 2] = int((fac[j, 1] + 1) * (fac[j, 1] + 2) / 2)
                else:
                    fac[:, 2] = 0  # no more combinations to try

    return p_size


def split_array(tab, psize):
    """ Split array into px*py*pz subarrays. """
    n_d = np.array(tab.shape, dtype=np.int64)
    slices = []
    for i in range(psize[0]):
        for j in range(psize[1]):
            for k in range(psize[2]):
                piece = np.array((i, j, k), dtype=np.int64)
                lei = n_d * piece / psize
                rei = n_d * (piece + np.ones(3, dtype=np.int64)) / psize
                slices.append(np.s_[lei[0]:rei[0], lei[1]:
                                    rei[1], lei[2]:rei[2]])
    return [tab[slc] for slc in slices]



"""
Enzo-specific IO functions

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2007-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from collections import defaultdict

try:
    from pyhdf_np import SD
except ImportError:
    pass

import exceptions
import os

from yt.utilities import hdf5_light_reader
from yt.utilities.io_handler import \
    BaseIOHandler, _axis_ids
from yt.utilities.logger import ytLogger as mylog

class IOHandlerEnzoHDF4(BaseIOHandler):

    _data_style = "enzo_hdf4"

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_field_names(self, grid):
        """
        Returns a list of fields associated with the filename
        Should *only* be called as EnzoGridInstance.getFields, never as getFields(object)
        """
        return SD.SD(grid.filename).datasets().keys()

    def _read_data(self, grid, field):
        """
        Returns after having obtained or generated a field.  Should throw an
        exception.  Should only be called as EnzoGridInstance.readData()

        @param field: field to read
        @type field: string
        """
        return SD.SD(grid.filename).select(field).get().swapaxes(0,2)

    @property
    def _read_exception(self):
        return SD.HDF4Error

class IOHandlerEnzoHDF5(BaseIOHandler):

    _data_style = "enzo_hdf5"
    _particle_reader = True

    def _read_field_names(self, grid):
        """
        Returns a list of fields associated with the filename
        Should *only* be called as EnzoGridInstance.getFields, never as getFields(object)
        """
        return hdf5_light_reader.ReadListOfDatasets(grid.filename, "/")

    def _read_data(self, grid, field):
        return hdf5_light_reader.ReadData(grid.filename, "/%s" % field).swapaxes(0,2)

    def modify(self, field):
        return field.swapaxes(0,2)

    @property
    def _read_exception(self):
        return (exceptions.KeyError, hdf5_light_reader.ReadingError)

    def _read_particles(self, fields, rtype, args, grid_list, enclosed,
                        conv_factors):
        filenames = [g.filename for g in grid_list]
        ids = [g.id for g in grid_list]
        return hdf5_light_reader.ReadParticles(
            rtype, fields, filenames, ids, conv_factors, args, 0)

class IOHandlerPackedHDF5(BaseIOHandler):

    _data_style = "enzo_packed_3d"
    _particle_reader = True

    def _read_particles(self, fields, rtype, args, grid_list, enclosed,
                        conv_factors):
        filenames = [g.filename for g in grid_list]
        ids = [g.id for g in grid_list]
        filenames, ids, conv_factors = zip(*sorted(zip(filenames, ids, conv_factors)))
        return hdf5_light_reader.ReadParticles(
            rtype, fields, list(filenames), list(ids), conv_factors, args, 1)

    def modify(self, field):
        return field.swapaxes(0,2)

    def preload(self, grids, sets):
        if len(grids) == 0:
            data = None
            return
        # We need to deal with files first
        files_keys = defaultdict(lambda: [])
        pf_field_list = grids[0].pf.h.field_list
        sets = [dset for dset in list(sets) if dset in pf_field_list]
        for g in grids:
            files_keys[g.filename].append(g)
        exc = self._read_exception
        for file in files_keys:
            # This is a funny business with Enzo files that are DM-only,
            # where grids can have *no* data, but still exist.
            if file is None: continue
            mylog.debug("Starting read %s (%s)", file, sets)
            nodes = [g.id for g in files_keys[file]]
            nodes.sort()
            # We want to pass on any error we might expect -- the preload
            # phase should be non-fatal in all cases, and instead dump back to
            # the grids.
            data = hdf5_light_reader.ReadMultipleGrids(file, nodes, sets)
            mylog.debug("Read %s items from %s", len(data), os.path.basename(file))
            for gid in data: self.queue[gid].update(data[gid])
        mylog.debug("Finished read of %s", sets)

    def _read_data(self, grid, field):
        tr = hdf5_light_reader.ReadData(grid.filename,
                "/Grid%08i/%s" % (grid.id, field))
        if tr.dtype == "float32": tr = tr.astype("float64")
        return self.modify(tr)

    def _read_field_names(self, grid):
        return hdf5_light_reader.ReadListOfDatasets(
                    grid.filename, "/Grid%08i" % grid.id)

    @property
    def _read_exception(self):
        return (exceptions.KeyError, hdf5_light_reader.ReadingError)

class IOHandlerPackedHDF5GhostZones(IOHandlerPackedHDF5):
    _data_style = "enzo_packed_3d_gz"

    def modify(self, field):
        if len(field.shape) < 3:
            return field
        tr = field[3:-3,3:-3,3:-3].swapaxes(0,2)
        return tr.copy() # To ensure contiguous

    def _read_raw_data_set(self, grid, field):
        return hdf5_light_reader.ReadData(grid.filename,
                "/Grid%08i/%s" % (grid.id, field))

class IOHandlerInMemory(BaseIOHandler):

    _data_style = "enzo_inline"

    def __init__(self, ghost_zones=3):
        import enzo
        self.enzo = enzo
        self.grids_in_memory = enzo.grid_data
        self.old_grids_in_memory = enzo.old_grid_data
        self.my_slice = (slice(ghost_zones,-ghost_zones),
                      slice(ghost_zones,-ghost_zones),
                      slice(ghost_zones,-ghost_zones))
        BaseIOHandler.__init__(self)

    def _read_data(self, grid, field):
        if grid.id not in self.grids_in_memory:
            mylog.error("Was asked for %s but I have %s", grid.id, self.grids_in_memory.keys())
            raise KeyError
        tr = self.grids_in_memory[grid.id][field]
        # If it's particles, we copy.
        if len(tr.shape) == 1: return tr.copy()
        # New in-place unit conversion breaks if we don't copy first
        return tr.swapaxes(0,2)[self.my_slice].copy()
        # We don't do this, because we currently do not interpolate
        coef1 = max((grid.Time - t1)/(grid.Time - t2), 0.0)
        coef2 = 1.0 - coef1
        t1 = enzo.yt_parameter_file["InitialTime"]
        t2 = enzo.hierarchy_information["GridOldTimes"][grid.id]
        return (coef1*self.grids_in_memory[grid.id][field] + \
                coef2*self.old_grids_in_memory[grid.id][field])\
                [self.my_slice]

    def modify(self, field):
        return field.swapaxes(0,2)

    def _read_field_names(self, grid):
        return self.grids_in_memory[grid.id].keys()

    def _read_data_slice(self, grid, field, axis, coord):
        # This data style cannot have a sidecar file
        sl = [slice(3,-3), slice(3,-3), slice(3,-3)]
        sl[axis] = slice(coord + 3, coord + 4)
        sl = tuple(reversed(sl))
        tr = self.grids_in_memory[grid.id][field][sl].swapaxes(0,2)
        # In-place unit conversion requires we return a copy
        return tr.copy()

    @property
    def _read_exception(self):
        return KeyError

class IOHandlerPacked2D(IOHandlerPackedHDF5):

    _data_style = "enzo_packed_2d"
    _particle_reader = False

    def _read_data(self, grid, field):
        return hdf5_light_reader.ReadData(grid.filename,
            "/Grid%08i/%s" % (grid.id, field)).transpose()[:,:,None]

    def modify(self, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        t = hdf5_light_reader.ReadData(grid.filename, "/Grid%08i/%s" %
                        (grid.id, field)).transpose()
        return t


class IOHandlerPacked1D(IOHandlerPackedHDF5):

    _data_style = "enzo_packed_1d"
    _particle_reader = False

    def _read_data(self, grid, field):
        return hdf5_light_reader.ReadData(grid.filename,
            "/Grid%08i/%s" % (grid.id, field)).transpose()[:,None,None]

    def modify(self, field):
        pass

    def _read_data_slice(self, grid, field, axis, coord):
        t = hdf5_light_reader.ReadData(grid.filename, "/Grid%08i/%s" %
                        (grid.id, field))
        return t




"""
Various definitions for various other modules and routines

Author: J. S. Oishi <jsoishi@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2010 J.S. Oishi.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
from yt.funcs import *

def boxlib_bool_to_int(v):
    try:
        return int(v)
    except ValueError:
        pass
    v = v.upper().strip()
    if v[0] == 'T':
        return 1
    elif v[0] == 'F':
        return 0

# TODO: get rid of enzo parameters we do not need
parameterDict = {"CosmologyCurrentRedshift": float,
                 "CosmologyComovingBoxSize": float,
                 "CosmologyOmegaMatterNow": float,
                 "CosmologyOmegaLambdaNow": float,
                 "CosmologyHubbleConstantNow": float,
                 "CosmologyInitialRedshift": float,
                 "DualEnergyFormalismEta1": float,
                 "DualEnergyFormalismEta2": float,
                 "MetaDataString": str,
                 "HydroMethod": int,
                 "DualEnergyFormalism": int,
                 "InitialTime": float,
                 "ComovingCoordinates": boxlib_bool_to_int,
                 "DensityUnits": float,
                 "LengthUnits": float,
                 "LengthUnit": float,
                 "TemperatureUnits": float,
                 "TimeUnits": float,
                 "GravitationalConstant": float,
                 "Gamma": float,
                 "MultiSpecies": int,
                 "CompilerPrecision": str,
                 "CurrentTimeIdentifier": int,
                 "RefineBy": int,
                 "BoundaryConditionName": str,
                 "TopGridRank": int,
                 "TopGridDimensions": int,
                 "EOSSoundSpeed": float,
                 "EOSType": int,
                 "NumberOfParticleAttributes": int,
                }

# converts the Castro inputs file name to the Enzo/yt name expected
# throughout the code. key is Castro name, value is Enzo/yt equivalent
castro2enzoDict = {"amr.n_cell": "TopGridDimensions",
                  "materials.gamma": "Gamma",
                  "amr.ref_ratio": "RefineBy",
                  "castro.use_comoving": "ComovingCoordinates",
                  "castro.redshift_in": "CosmologyInitialRedshift",
                  "comoving_OmL": "CosmologyOmegaLambdaNow",
                  "comoving_OmM": "CosmologyOmegaMatterNow",
                  "comoving_h": "CosmologyHubbleConstantNow"
                  }

yt2castroFieldsDict = {}
castro2ytFieldsDict = {}

castro_FAB_header_pattern = r"^FAB \(\((\d+), \([0-9 ]+\)\),\(\d+, \(([0-9 ]+)\)\)\)\(\((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\) \((\d+,\d+,\d+)\)\) (\d+)\n"

castro_particle_field_names = \
    ['particle_position_%s' % ax for ax in 'xyz'] + \
    ['particle_mass'] +  \
    ['particle_velocity_%s' % ax for ax in 'xyz']



"""
Clump finding helper classes

Author: Britton Smith <Britton.Smith@colorado.edu>
Affiliation: University of Colorado at Boulder
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import copy

from yt.funcs import *

from .contour_finder import identify_contours

class Clump(object):
    children = None
    def __init__(self, data, parent, field, cached_fields = None, 
                 function=None, clump_info=None):
        self.parent = parent
        self.data = data
        self.quantities = data.quantities
        self.field = field
        self.min_val = self.data[field].min()
        self.max_val = self.data[field].max()
        self.cached_fields = cached_fields

        # List containing characteristics about clumps that are to be written 
        # out by the write routines.
        if clump_info is None:
            self.set_default_clump_info()
        else:
            # Clump info will act the same if add_info_item is called before or after clump finding.
            self.clump_info = copy.deepcopy(clump_info)

        # Function determining whether a clump is valid and should be kept.
        self.default_function = 'self.data.quantities["IsBound"](truncate=True,include_thermal_energy=True) > 1.0'
        if function is None:
            self.function = self.default_function
        else:
            self.function = function

        # Return value of validity function, saved so it does not have to be calculated again.
        self.function_value = None

    def add_info_item(self,quantity,format):
        "Adds an entry to clump_info list and tells children to do the same."

        self.clump_info.append({'quantity':quantity, 'format':format})
        if self.children is None: return
        for child in self.children:
            child.add_info_item(quantity,format)

    def set_default_clump_info(self):
        "Defines default entries in the clump_info array."

        # add_info_item is recursive so this function does not need to be.
        self.clump_info = []

        # Number of cells.
        self.add_info_item('self.data["CellMassMsun"].size','"Cells: %d" % value')
        # Gas mass in solar masses.
        self.add_info_item('self.data["CellMassMsun"].sum()','"Mass: %e Msolar" % value')
        # Volume-weighted Jeans mass.
        self.add_info_item('self.data.quantities["WeightedAverageQuantity"]("JeansMassMsun","CellVolume")',
                           '"Jeans Mass (vol-weighted): %.6e Msolar" % value')
        # Mass-weighted Jeans mass.
        self.add_info_item('self.data.quantities["WeightedAverageQuantity"]("JeansMassMsun","CellMassMsun")',
                           '"Jeans Mass (mass-weighted): %.6e Msolar" % value')
        # Max level.
        self.add_info_item('self.data["GridLevel"].max()','"Max grid level: %d" % value')
        # Minimum number density.
        self.add_info_item('self.data["NumberDensity"].min()','"Min number density: %.6e cm^-3" % value')
        # Maximum number density.
        self.add_info_item('self.data["NumberDensity"].max()','"Max number density: %.6e cm^-3" % value')

    def clear_clump_info(self):
        "Clears the clump_info array and passes the instruction to its children."

        self.clump_info = []
        if self.children is None: return
        for child in self.children:
            child.clear_clump_info()

    def write_info(self,level,f_ptr):
        "Writes information for clump using the list of items in clump_info."

        for item in self.clump_info:
            # Call if callable, otherwise do an eval.
            if callable(item['quantity']):
                value = item['quantity']()
            else:
                value = eval(item['quantity'])
            output = eval(item['format'])
            f_ptr.write("%s%s" % ('\t'*level,output))
            f_ptr.write("\n")

    def find_children(self, min_val, max_val = None):
        if self.children is not None:
            print "Wiping out existing children clumps."
        self.children = []
        if max_val is None: max_val = self.max_val
        contour_info = identify_contours(self.data, self.field, min_val, max_val,
                                         self.cached_fields)
        for cid in contour_info:
            new_clump = self.data.extract_region(contour_info[cid])
            self.children.append(Clump(new_clump, self, self.field,
                                       self.cached_fields,function=self.function,
                                       clump_info=self.clump_info))

    def pass_down(self,operation):
        "Performs an operation on a clump with an exec and passes the instruction down to clump children."

        # Call if callable, otherwise do an exec.
        if callable(operation):
            operation()
        else:
            exec(operation)

        if self.children is None: return
        for child in self.children:
            child.pass_down(operation)

    def _isValid(self):
        "Perform user specified function to determine if child clumps should be kept."

        # Only call function if it has not been already.
        if self.function_value is None:
            self.function_value = eval(self.function)

        return self.function_value

    def __reduce__(self):
        return (_reconstruct_clump, 
                (self.parent, self.field, self.min_val, self.max_val,
                 self.function_value, self.children, self.data, self.clump_info, self.function))

    def __getitem__(self,request):
        return self.data[request]

def _reconstruct_clump(parent, field, mi, ma, function_value, children, data, clump_info, 
        function=None):
    obj = object.__new__(Clump)
    if iterable(parent):
        try:
            parent = parent[1]
        except KeyError:
            parent = parent
    if children is None: children = []
    obj.parent, obj.field, obj.min_val, obj.max_val, obj.function_value, obj.children, obj.clump_info, obj.function = \
        parent, field, mi, ma, function_value, children, clump_info, function
    # Now we override, because the parent/child relationship seems a bit
    # unreliable in the unpickling
    for child in children: child.parent = obj
    obj.data = data[1] # Strip out the PF
    obj.quantities = obj.data.quantities
    if obj.parent is None: return (data[0], obj)
    return obj

def find_clumps(clump, min_val, max_val, d_clump):
    print "Finding clumps: min: %e, max: %e, step: %f" % (min_val, max_val, d_clump)
    if min_val >= max_val: return
    clump.find_children(min_val)

    if (len(clump.children) == 1):
        find_clumps(clump, min_val*d_clump, max_val, d_clump)

    elif (len(clump.children) > 0):
        these_children = []
        print "Investigating %d children." % len(clump.children)
        for child in clump.children:
            find_clumps(child, min_val*d_clump, max_val, d_clump)
            if ((child.children is not None) and (len(child.children) > 0)):
                these_children.append(child)
            elif (child._isValid()):
                these_children.append(child)
            else:
                print "Eliminating invalid, childless clump with %d cells." % len(child.data["Ones"])
        if (len(these_children) > 1):
            print "%d of %d children survived." % (len(these_children),len(clump.children))            
            clump.children = these_children
        elif (len(these_children) == 1):
            print "%d of %d children survived, linking its children to parent." % (len(these_children),len(clump.children))
            clump.children = these_children[0].children
        else:
            print "%d of %d children survived, erasing children." % (len(these_children),len(clump.children))
            clump.children = []

def get_lowest_clumps(clump, clump_list=None):
    "Return a list of all clumps at the bottom of the hierarchy."

    if clump_list is None: clump_list = []
    if clump.children is None or len(clump.children) == 0:
        clump_list.append(clump)
    if clump.children is not None and len(clump.children) > 0:
        for child in clump.children:
            get_lowest_clumps(child, clump_list=clump_list)

    return clump_list

def write_clump_hierarchy(clump,level,f_ptr):
    for q in range(level):
        f_ptr.write("\t")
    f_ptr.write("Clump at level %d:\n" % level)
    clump.write_info(level,f_ptr)
    f_ptr.write("\n")
    f_ptr.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clump_hierarchy(child,(level+1),f_ptr)

def write_clumps(clump,level,f_ptr):
    if ((clump.children is None) or (len(clump.children) == 0)):
        f_ptr.write("%sClump:\n" % ("\t"*level))
        clump.write_info(level,f_ptr)
        f_ptr.write("\n")
        f_ptr.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clumps(child,0,f_ptr)

# Old clump info writing routines.
def write_old_clump_hierarchy(clump,level,f_ptr):
    for q in range(level):
        f_ptr.write("\t")
    f_ptr.write("Clump at level %d:\n" % level)
    clump.write_info(level,f_ptr)
    write_old_clump_info(clump,level,f_ptr)
    f_ptr.write("\n")
    f_ptr.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clump_hierarchy(child,(level+1),f_ptr)

def write_old_clumps(clump,level,f_ptr):
    if ((clump.children is None) or (len(clump.children) == 0)):
        f_ptr.write("%sClump:\n" % ("\t"*level))
        write_old_clump_info(clump,level,f_ptr)
        f_ptr.write("\n")
        f_ptr.flush()
    if ((clump.children is not None) and (len(clump.children) > 0)):
        for child in clump.children:
            write_clumps(child,0,f_ptr)

__clump_info_template = \
"""
%(tl)sCells: %(num_cells)s
%(tl)sMass: %(total_mass).6e Msolar
%(tl)sJeans Mass (vol-weighted): %(jeans_mass_vol).6e Msolar
%(tl)sJeans Mass (mass-weighted): %(jeans_mass_mass).6e Msolar
%(tl)sMax grid level: %(max_level)s
%(tl)sMin number density: %(min_density).6e cm^-3
%(tl)sMax number density: %(max_density).6e cm^-3

"""

def write_old_clump_info(clump,level,f_ptr):
    fmt_dict = {'tl':  "\t" * level}
    fmt_dict['num_cells'] = clump.data["CellMassMsun"].size,
    fmt_dict['total_mass'] = clump.data["CellMassMsun"].sum()
    fmt_dict['jeans_mass_vol'] = clump.data.quantities["WeightedAverageQuantity"]("JeansMassMsun","CellVolume")
    fmt_dict['jeans_mass_mass'] = clump.data.quantities["WeightedAverageQuantity"]("JeansMassMsun","CellMassMsun")
    fmt_dict['max_level'] =  clump.data["GridLevel"].max()
    fmt_dict['min_density'] =  clump.data["NumberDensity"].min()
    fmt_dict['max_density'] =  clump.data["NumberDensity"].max()
    f_ptr.write(__clump_info_template % fmt_dict)

# Recipes for various clump calculations.
recipes = {}

# Distance from clump center of mass to center of mass of top level object.
def _DistanceToMainClump(master,units='pc'):
    masterCOM = master.data.quantities['CenterOfMass']()
    pass_command = "self.masterCOM = [%.10f, %.10f, %.10f]" % (masterCOM[0],
                                                               masterCOM[1],
                                                               masterCOM[2])
    master.pass_down(pass_command)
    master.pass_down("self.com = self.data.quantities['CenterOfMass']()")

    quantity = "((self.com[0]-self.masterCOM[0])**2 + (self.com[1]-self.masterCOM[1])**2 + (self.com[2]-self.masterCOM[2])**2)**(0.5)*self.data.pf.units['%s']" % units
    format = "%s%s%s" % ("'Distance from center: %.6e ",units,"' % value")

    master.add_info_item(quantity,format)

recipes['DistanceToMainClump'] = _DistanceToMainClump



"""
An attempt at putting the ray-casting operation into CUDA

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2009 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys;sys.path.insert(0,'.')

from yt.mods import *
import yt.extensions.HierarchySubset as hs
import numpy as np
import h5py, time

import matplotlib;matplotlib.use("Agg");import pylab

from yt.extensions.volume_rendering.TransferFunction import ColorTransferFunction

if __name__ == "__main__":

    # This is boilerplate code for setting up pycuda
    import pycuda.driver as cuda
    import pycuda.compiler as compiler
    import pycuda.autoinit
    import pycuda.gpuarray as gpuarray
    cuda.init()
    assert (cuda.Device.count() >= 1)

    print "Extracting hierarchy."
    opf = load("/u/ki/mturk/ki05/MSM96-SIM3-restart-J64/DataDump0081.dir/DataDump0081")
    pf = hs.ExtractedParameterFile(opf, 20)

    cpu = {}
    gpu = {}

    print "Reading data."
    #fn = "DataDump0081_partitioned.h5"
    fn = "RedshiftOutput0005_partitioned.h5"
    f = h5py.File("/u/ki/mturk/ki05/%s" % fn)
    cpu['grid_data'] = f["/PGrids/Data"][:].astype("float32")
    cpu['dims'] = f["/PGrids/Dims"][:].astype("int32") - 1
    cpu['left_edge'] = f["/PGrids/LeftEdges"][:].astype("float32")
    cpu['right_edge'] = f["/PGrids/RightEdges"][:].astype("float32")

    print "Constructing transfer function."
    if "Data" in fn:
        mh = np.log10(1.67e-24)
        tf = ColorTransferFunction((7.5+mh, 14.0+mh))
        tf.add_gaussian( 8.25+mh, 0.002, [0.2, 0.2, 0.4, 0.1])
        tf.add_gaussian( 9.75+mh, 0.002, [0.0, 0.0, 0.3, 0.1])
        tf.add_gaussian(10.25+mh, 0.004, [0.0, 0.3, 0.0, 0.1])
        tf.add_gaussian(11.50+mh, 0.005, [1.0, 0.0, 0.0, 0.7])
        tf.add_gaussian(12.75+mh, 0.010, [1.0, 1.0, 1.0, 1.0])
    elif "Red" in fn:
        tf = ColorTransferFunction((-31, -27))
        tf.add_gaussian(-30.0, 0.05, [1.0, 0.0, 0.0, 0.1])
        tf.add_gaussian(-29.5, 0.03, [0.0, 1.0, 0.0, 0.3])
        tf.add_gaussian(-29.0, 0.05, [0.0, 0.0, 1.0, 0.5])
        tf.add_gaussian(-28.5, 0.05, [1.0, 1.0, 1.0, 1.0])
    else: raise RuntimeError

    cpu['ngrids'] = np.array([cpu['dims'].shape[0]], dtype='int32')
    cpu['tf_r'] = tf.red.y.astype("float32")
    cpu['tf_g'] = tf.green.y.astype("float32")
    cpu['tf_b'] = tf.blue.y.astype("float32")
    cpu['tf_a'] = tf.alpha.y.astype("float32")

    cpu['tf_bounds'] = np.array(tf.x_bounds, dtype='float32')

    cpu['v_dir'] = np.array([0.3, 0.5, 0.6], dtype='float32')

    c = np.array([0.47284317, 0.48062515, 0.58282089], dtype='float32')

    print "Getting cutting plane."
    cp = pf.h.cutting(cpu['v_dir'], c)

    W = 2000.0/pf['au']
    W = 0.25
    Nvec = 128
    back_c = c - cp._norm_vec * W
    front_c = c + cp._norm_vec * W

    px, py = np.mgrid[-W:W:Nvec*1j,-W:W:Nvec*1j]
    xv = cp._inv_mat[0,0]*px + cp._inv_mat[0,1]*py + cp.center[0]
    yv = cp._inv_mat[1,0]*px + cp._inv_mat[1,1]*py + cp.center[1]
    zv = cp._inv_mat[2,0]*px + cp._inv_mat[2,1]*py + cp.center[2]
    cpu['v_pos'] = np.array([xv, yv, zv], dtype='float32').transpose()

    cpu['image_r'] = np.zeros((Nvec, Nvec), dtype='float32').ravel()
    cpu['image_g'] = np.zeros((Nvec, Nvec), dtype='float32').ravel()
    cpu['image_b'] = np.zeros((Nvec, Nvec), dtype='float32').ravel()
    cpu['image_a'] = np.zeros((Nvec, Nvec), dtype='float32').ravel()

    print "Generating module"
    source = open("yt/extensions/volume_rendering/_cuda_caster.cu").read()
    mod = compiler.SourceModule(source)
    func = mod.get_function("ray_cast")

    for n, a in cpu.items():
        ss = a.size * a.dtype.itemsize
        print "Allocating %0.3e megabytes for %s" % (ss/(1024*1024.), n)
        gpu[n] = cuda.to_device(a.ravel('F'))
        #pycuda.autoinit.context.synchronize()

    BLOCK_SIZE = 8
    grid_size = Nvec / BLOCK_SIZE

    print "Running ray_cast function."
    t1 = time.time()
    ret = func(gpu['ngrids'],
               gpu['grid_data'],
               gpu['dims'],
               gpu['left_edge'],
               gpu['right_edge'],
               gpu['tf_r'],
               gpu['tf_g'],
               gpu['tf_b'],
               gpu['tf_a'],
               gpu['tf_bounds'],
               gpu['v_dir'],
               gpu['v_pos'],
               gpu['image_r'],
               gpu['image_g'],
               gpu['image_b'],
               gpu['image_a'],
         block=(BLOCK_SIZE,BLOCK_SIZE,1),
         grid=(grid_size, grid_size), time_kernel=True)
    t2 = time.time()
    print "BACK: %0.3e" % (t2-t1)

    mi, ma = 1e300, -1e300
    image = []
    for im in 'rgb':
        ii = 'image_%s' % im
        sh, dtype = cpu[ii].shape, cpu[ii].dtype
        del cpu[ii]
        cpu[ii] = cuda.from_device(gpu[ii], sh, dtype).reshape((Nvec,Nvec))
        mi, ma = min(cpu[ii].min(),mi), max(cpu[ii].max(), ma)
        image.append(cpu[ii])
        print "Min/max of %s %0.3e %0.3e" % (
                im, image[-1].min(), image[-1].max())
        pylab.clf()
        pylab.imshow(image[-1], interpolation='nearest')
        pylab.savefig("/u/ki/mturk/public_html/vr6/%s.png" % (ii))

    image = np.array(image).transpose()
    image = (image - mi) / (ma - mi)
    pylab.clf()
    pylab.imshow(image, interpolation='nearest')
    pylab.savefig("/u/ki/mturk/public_html/vr6/image_rgb.png")



"""
Create randomly centered, tiled projections to be used in light cones.

Author: Britton Smith <brittons@origins.colorado.edu>
Affiliation: CASA/University of CO, Boulder
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Britton Smith.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import copy

from yt.funcs import *
from yt.visualization.fixed_resolution import \
    FixedResolutionBuffer
from yt.utilities.parallel_tools.parallel_analysis_interface import \
    parallel_blocking_call

@parallel_blocking_call
def _light_cone_projection(lightConeSlice, field, pixels, weight_field=None,
                           save_image=False, node=None, field_cuts=None):
    "Create a single projection to be added into the light cone stack."

    # Use some projection parameters to seed random number generator to make unique node name.
    # We are just saving the projection object, so only the projection axis needs to be considered
    # since the lateral shifting and tiling occurs after the projection object is made.
    # Likewise, only the box_depth_fraction needs to be considered.

    # projection_axis
    # projection_center[projection_axis]
    # box_depth_fraction

    # Name node with user specified keyword if given with 'node' keyword.
    node_name = "LightCone_%s_%d_%f_%f" % (node, lightConeSlice['projection_axis'],
                                           lightConeSlice['projection_center'][lightConeSlice['projection_axis']],
                                           lightConeSlice['box_depth_fraction'])

    mylog.info("Making projection at z = %f from %s." % (lightConeSlice['redshift'], lightConeSlice['filename']))

    region_center = [0.5 * (lightConeSlice['object'].domain_right_edge[q] +
                            lightConeSlice['object'].domain_left_edge[q]) \
                         for q in range(lightConeSlice['object'].dimensionality)]

    # 1. The Depth Problem
    # Use coordinate field cut in line of sight to cut projection to proper depth.
    if field_cuts is None:
        these_field_cuts = []
    else:
        these_field_cuts = copy.deepcopy(field_cuts)

    if (lightConeSlice['box_depth_fraction'] < 1):
        axis = ('x', 'y', 'z')[lightConeSlice['projection_axis']]
        depthLeft = lightConeSlice['projection_center'][lightConeSlice['projection_axis']] \
            - 0.5 * lightConeSlice['box_depth_fraction']
        depthRight = lightConeSlice['projection_center'][lightConeSlice['projection_axis']] \
            + 0.5 * lightConeSlice['box_depth_fraction']
        if (depthLeft < 0):
            cut_mask = "((grid[\"%s\"] + 0.5*grid[\"d%s\"] >= 0) & (grid[\"%s\"] - 0.5*grid[\"d%s\"] <= %f)) | ((grid[\"%s\"] + 0.5*grid[\"d%s\"] >= %f) & (grid[\"%s\"] - 0.5*grid[\"d%s\"] <= 1))" % \
                (axis, axis, axis, axis, depthRight, axis, axis, (depthLeft+1), axis, axis)
        elif (depthRight > 1):
            cut_mask = "((grid[\"%s\"] + 0.5*grid[\"d%s\"] >= 0) & (grid[\"%s\"] - 0.5*grid[\"d%s\"] <= %f)) | ((grid[\"%s\"] + 0.5*grid[\"d%s\"] >= %f) & (grid[\"%s\"] - 0.5*grid[\"d%s\"] <= 1))" % \
                (axis, axis, axis, axis, (depthRight-1), axis, axis, depthLeft, axis, axis)
        else:
            cut_mask = "(grid[\"%s\"] + 0.5*grid[\"d%s\"] >= %f) & (grid[\"%s\"] - 0.5*grid[\"%s\"] <= %f)" % (axis, axis, depthLeft, axis, axis, depthRight)

        these_field_cuts.append(cut_mask)

    # Make projection.
    proj = lightConeSlice['object'].h.proj(lightConeSlice['projection_axis'], field,
                                           weight_field, center=region_center,
                                           field_cuts=these_field_cuts, node_name=node_name)

    # 2. The Tile Problem
    # Tile projection to specified width.

    # Original projection data.
    original_px = copy.deepcopy(proj['px'])
    original_py = copy.deepcopy(proj['py'])
    original_pdx = copy.deepcopy(proj['pdx'])
    original_pdy = copy.deepcopy(proj['pdy'])
    original_field = copy.deepcopy(proj[field])
    original_weight_field = copy.deepcopy(proj['weight_field'])

    # Copy original into offset positions to make tiles.
    for x in range(int(np.ceil(lightConeSlice['box_width_fraction']))):
        for y in range(int(np.ceil(lightConeSlice['box_width_fraction']))):
            if ((x + y) > 0):
                proj['px'] = np.concatenate([proj['px'], original_px+x])
                proj['py'] = np.concatenate([proj['py'], original_py+y])
                proj['pdx'] = np.concatenate([proj['pdx'], original_pdx])
                proj['pdy'] = np.concatenate([proj['pdy'], original_pdy])
                proj[field] = np.concatenate([proj[field], original_field])
                proj['weight_field'] = np.concatenate([proj['weight_field'],
                                                       original_weight_field])

    # Delete originals.
    del original_px
    del original_py
    del original_pdx
    del original_pdy
    del original_field
    del original_weight_field

    # 3. The Shift Problem
    # Shift projection by random x and y offsets.

    offset = copy.deepcopy(lightConeSlice['projection_center'])
    # Delete depth coordinate.
    del offset[lightConeSlice['projection_axis']]

    # Shift x and y positions.
    proj['px'] -= offset[0]
    proj['py'] -= offset[1]

    # Wrap off-edge cells back around to other side (periodic boundary conditions).
    proj['px'][proj['px'] < 0] += np.ceil(lightConeSlice['box_width_fraction'])
    proj['py'][proj['py'] < 0] += np.ceil(lightConeSlice['box_width_fraction'])

    # After shifting, some cells have fractional coverage on both sides of the box.
    # Find those cells and make copies to be placed on the other side.

    # Cells hanging off the right edge.
    add_x_right = proj['px'] + 0.5 * proj['pdx'] > \
      np.ceil(lightConeSlice['box_width_fraction'])
    add_x_px = proj['px'][add_x_right]
    add_x_px -= np.ceil(lightConeSlice['box_width_fraction'])
    add_x_py = proj['py'][add_x_right]
    add_x_pdx = proj['pdx'][add_x_right]
    add_x_pdy = proj['pdy'][add_x_right]
    add_x_field = proj[field][add_x_right]
    add_x_weight_field = proj['weight_field'][add_x_right]
    del add_x_right

    # Cells hanging off the left edge.
    add_x_left = proj['px'] - 0.5 * proj['pdx'] < 0
    add2_x_px = proj['px'][add_x_left]
    add2_x_px += np.ceil(lightConeSlice['box_width_fraction'])
    add2_x_py = proj['py'][add_x_left]
    add2_x_pdx = proj['pdx'][add_x_left]
    add2_x_pdy = proj['pdy'][add_x_left]
    add2_x_field = proj[field][add_x_left]
    add2_x_weight_field = proj['weight_field'][add_x_left]
    del add_x_left

    # Cells hanging off the top edge.
    add_y_right = proj['py'] + 0.5 * proj['pdy'] > \
      np.ceil(lightConeSlice['box_width_fraction'])
    add_y_px = proj['px'][add_y_right]
    add_y_py = proj['py'][add_y_right]
    add_y_py -= np.ceil(lightConeSlice['box_width_fraction'])
    add_y_pdx = proj['pdx'][add_y_right]
    add_y_pdy = proj['pdy'][add_y_right]
    add_y_field = proj[field][add_y_right]
    add_y_weight_field = proj['weight_field'][add_y_right]
    del add_y_right

    # Cells hanging off the bottom edge.
    add_y_left = proj['py'] - 0.5 * proj['pdy'] < 0
    add2_y_px = proj['px'][add_y_left]
    add2_y_py = proj['py'][add_y_left]
    add2_y_py += np.ceil(lightConeSlice['box_width_fraction'])
    add2_y_pdx = proj['pdx'][add_y_left]
    add2_y_pdy = proj['pdy'][add_y_left]
    add2_y_field = proj[field][add_y_left]
    add2_y_weight_field = proj['weight_field'][add_y_left]
    del add_y_left

    # Add the hanging cells back to the projection data.
    proj['px'] = np.concatenate([proj['px'], add_x_px, add_y_px,
                                 add2_x_px, add2_y_px])
    proj['py'] = np.concatenate([proj['py'], add_x_py, add_y_py,
                                 add2_x_py, add2_y_py])
    proj['pdx'] = np.concatenate([proj['pdx'], add_x_pdx, add_y_pdx,
                                  add2_x_pdx, add2_y_pdx])
    proj['pdy'] = np.concatenate([proj['pdy'], add_x_pdy, add_y_pdy,
                                  add2_x_pdy, add2_y_pdy])
    proj[field] = np.concatenate([proj[field], add_x_field, add_y_field,
                                  add2_x_field, add2_y_field])
    proj['weight_field'] = np.concatenate([proj['weight_field'],
                                           add_x_weight_field, add_y_weight_field,
                                           add2_x_weight_field, add2_y_weight_field])

    # Delete original copies of hanging cells.
    del add_x_px, add_y_px, add2_x_px, add2_y_px
    del add_x_py, add_y_py, add2_x_py, add2_y_py
    del add_x_pdx, add_y_pdx, add2_x_pdx, add2_y_pdx
    del add_x_pdy, add_y_pdy, add2_x_pdy, add2_y_pdy
    del add_x_field, add_y_field, add2_x_field, add2_y_field
    del add_x_weight_field, add_y_weight_field, add2_x_weight_field, add2_y_weight_field

    # Tiles were made rounding up the width to the nearest integer.
    # Cut off the edges to get the specified width.
    # Cut in the x direction.
    cut_x = proj['px'] - 0.5 * proj['pdx'] < lightConeSlice['box_width_fraction']
    proj['px'] = proj['px'][cut_x]
    proj['py'] = proj['py'][cut_x]
    proj['pdx'] = proj['pdx'][cut_x]
    proj['pdy'] = proj['pdy'][cut_x]
    proj[field] = proj[field][cut_x]
    proj['weight_field'] = proj['weight_field'][cut_x]
    del cut_x

    # Cut in the y direction.
    cut_y = proj['py'] - 0.5 * proj['pdy'] < lightConeSlice['box_width_fraction']
    proj['px'] = proj['px'][cut_y]
    proj['py'] = proj['py'][cut_y]
    proj['pdx'] = proj['pdx'][cut_y]
    proj['pdy'] = proj['pdy'][cut_y]
    proj[field] = proj[field][cut_y]
    proj['weight_field'] = proj['weight_field'][cut_y]
    del cut_y

    # Create fixed resolution buffer to return back to the light cone object.
    # These buffers will be stacked together to make the light cone.
    frb = FixedResolutionBuffer(proj, (0, lightConeSlice['box_width_fraction'],
                                       0, lightConeSlice['box_width_fraction']),
                                (pixels, pixels), antialias=False)

    return frb



"""
A means of extracting a subset of the hierarchy

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008-2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import h5py, os.path
import numpy as np

from yt.funcs import *
from yt.data_objects.data_containers import YTFieldData
from yt.data_objects.grid_patch import \
    AMRGridPatch
from yt.data_objects.static_output import \
    StaticOutput
from yt.data_objects.hierarchy import \
    AMRHierarchy

class DummyHierarchy(object):
    pass

class ConstructedRootGrid(AMRGridPatch):
    __slots__ = ['base_grid', 'id', 'base_pf']
    _id_offset = 1
    def __init__(self, base_pf, pf, hierarchy, level, left_edge, right_edge):
        """
        This is a fake root grid, constructed by creating a
        :class:`yt.data_objects.api.CoveringGridBase` at a given *level* between
        *left_edge* and *right_edge*.
        """
        self.pf = pf
        self.base_pf = base_pf
        self.field_parameters = {}
        self.NumberOfParticles = 0
        self.id = 1
        self.hierarchy = hierarchy
        self._child_mask = self._child_indices = self._child_index_mask = None
        self.Level = level
        self.LeftEdge = left_edge
        self.RightEdge = right_edge
        self.start_index = np.min([grid.get_global_startindex() for grid in
                             base_pf.h.select_grids(level)], axis=0).astype('int64')
        self.dds = base_pf.h.select_grids(level)[0].dds.copy()
        dims = (self.RightEdge-self.LeftEdge)/self.dds
        self.ActiveDimensions = dims
        print "Constructing base grid of size %s" % (self.ActiveDimensions)
        self.base_grid = base_pf.h.smoothed_covering_grid(level, self.LeftEdge,
                        self.RightEdge, dims=dims)
        self.base_grid.Level = self.base_grid.level
        self.field_data = YTFieldData()
        #self._calculate_child_masks()
        self.Parent = None
        self.Children = []

    def get_vertex_centered_data(self, field, smoothed=True):
        vc = self.base_pf.h.smoothed_covering_grid(self.base_grid.Level,
                self.base_grid.LeftEdge - self.base_grid.dds*0.5,
                self.base_grid.RightEdge + self.base_grid.dds*0.5,
                dims = self.ActiveDimensions + 1)
        return vc[field]

class AMRExtractedGridProxy(AMRGridPatch):
    __slots__ = ['base_grid']
    _id_offset = 1
    def __init__(self, grid_id, base_grid, hierarchy):
        # We make a little birdhouse in our soul for the base_grid
        # (they're the only bee in our bonnet!)
        self.base_grid = base_grid
        AMRGridPatch.__init__(self, grid_id, filename = None, hierarchy=hierarchy)
        self.Parent = None
        self.Children = []
        self.Level = -1

    def get_vertex_centered_data(self, *args, **kwargs):
        return self.base_grid.get_vertex_centered_data(*args, **kwargs)

class OldExtractedHierarchy(object):

    def __init__(self, pf, min_level, max_level = -1, offset = None,
                 always_copy=False):
        """
        This is a class that extracts a hierarchy from another hierarchy,
        filling in regions as necessary.  It accepts a parameter file (*pf*), a
        *min_level*, a *max_level*, and alternately an *offset*.  This class is
        typically or exclusively used to extract for the purposes of visualization.
        """
        self.pf = pf
        self.always_copy = always_copy
        self.min_level = min_level
        self.int_offset = np.min([grid.get_global_startindex() for grid in
                             pf.h.select_grids(min_level)], axis=0).astype('float64')
        min_left = np.min([grid.LeftEdge for grid in
                           pf.h.select_grids(min_level)], axis=0).astype('float64')
        max_right = np.max([grid.RightEdge for grid in 
                                   pf.h.select_grids(min_level)], axis=0).astype('float64')
        if offset is None: offset = (max_right + min_left)/2.0
        self.left_edge_offset = offset
        self.mult_factor = 2**min_level
        self.min_left_edge = self._convert_coords(min_left)
        self.max_right_edge = self._convert_coords(max_right)
        if max_level == -1: max_level = pf.h.max_level
        self.max_level = min(max_level, pf.h.max_level)
        self.final_level = self.max_level - self.min_level
        if len(self.pf.h.select_grids(self.min_level)) > 0:
            self._base_grid = ConstructedRootGrid(self.pf, self.min_level,
                               min_left, max_right)
        else: self._base_grid = None
        
    def select_level(self, level):
        if level == 0 and self._base_grid is not None:
            return [self._base_grid]
        return self.pf.h.select_grids(self.min_level + level)

    def export_output(self, afile, n, field):
        # I prefer dict access, but tables doesn't.
        # But h5py does!
        time_node = afile.create_group("/time-%s" % n)
        time_node.attrs['time'] = self.pf.current_time
        time_node.attrs['numLevels'] = self.pf.h.max_level+1-self.min_level
        # Can take a while, so let's get a progressbar
        self._export_all_levels(afile, time_node, field)

    def _export_all_levels(self, afile, time_node, field):
        pbar = yt.funcs.get_pbar("Exporting levels", self.final_level+1)
        for i,grid_set in enumerate(self.get_levels()):
            pbar.update(i)
            self.export_level(afile, time_node, i, field, grid_set)
        pbar.finish()

    def export_level(self, afile, time_node, level, field, grids = None):
        level_node = afile.create_group("%s/level-%s" % (time_node,level))
        # Grid objects on this level...
        if grids is None: grids = self.pf.h.select_grids(level+self.min_level)
        level_node.attrs['delta'] = grids[0].dds*self.mult_factor
        level_node.attrs['relativeRefinementFactor'] = np.array([2]*3, dtype='int32')
        level_node.attrs['numGrids'] = len(grids)
        for i,g in enumerate(grids):
            self.export_grid(afile, level_node, g, i, field)

    def _convert_grid(self, grid):
        int_origin = (grid.get_global_startindex() \
                    - self.int_offset*2**(grid.Level-self.min_level)).astype('int64')
        level_int_origin = (grid.LeftEdge - self.left_edge_offset)/grid.dds
        origin = self._convert_coords(grid.LeftEdge)
        dds = grid.dds * self.mult_factor
        return int_origin, level_int_origin, origin, dds

    def export_grid(self, afile, level_node, grid, i, field):
        grid_node = afile.create_group("%s/grid-%s" % (level_node,i))
        int_origin, lint, origin, dds = self._convert_grid(grid)
        grid_node.attrs['integerOrigin'] = int_origin
        grid_node.attrs['origin'] = origin
        grid_node.attrs['ghostzoneFlags'] = np.zeros(6, dtype='int32')
        grid_node.attrs['numGhostzones'] = np.zeros(3, dtype='int32')
        grid_node.attrs['dims'] = grid.ActiveDimensions[::-1].astype('int32')
        if not self.always_copy and self.pf.h.data_style == 6 \
           and field in self.pf.h.field_list:
            if grid.hierarchy.data_style == -1: # constructed grid
                # if we can get conversion in amira we won't need to do this
                ff = grid[field].astype('float32')
                ff /= self.pf.conversion_factors.get(field, 1.0)
                afile.create_dataset("%s/grid-data" % grid_node, data=ff.swapaxes(0,2))
            else:
                tfn = os.path.abspath(afile.filename)
                gfn = os.path.abspath(grid.filename)
                fpn = os.path.commonprefix([tfn, grid.filename])
                fn = grid.filename[len(os.path.commonprefix([tfn, grid.filename])):]
                grid_node.attrs['referenceFileName'] = fn
                grid_node.attrs['referenceDataPath'] = \
                    "/Grid%08i/%s" % (grid.id, field)
        else:
            # Export our array
            afile.create_dataset("%s/grid-data" % grid_node, 
                                 data = grid[field].astype('float32').swapaxes(0,2))

    def _convert_coords(self, val):
        return (val - self.left_edge_offset)*self.mult_factor

class ExtractedHierarchy(AMRHierarchy):

    grid = AMRExtractedGridProxy

    def __init__(self, pf, data_style):
        # First we set up our translation between original and extracted
        self.data_style = data_style
        self.min_level = pf.min_level
        self.int_offset = np.min([grid.get_global_startindex() for grid in
                           pf.base_pf.h.select_grids(pf.min_level)], axis=0).astype('float64')
        min_left = np.min([grid.LeftEdge for grid in
                           pf.base_pf.h.select_grids(pf.min_level)], axis=0).astype('float64')
        max_right = np.max([grid.RightEdge for grid in 
                           pf.base_pf.h.select_grids(pf.min_level)], axis=0).astype('float64')
        level_dx = pf.base_pf.h.select_grids(pf.min_level)[0].dds[0]
        dims = ((max_right-min_left)/level_dx)
        max_right += (dims.max() - dims) * level_dx
        offset = pf.offset
        if offset is None: offset = min_left
        self.left_edge_offset = offset
        pf.offset = offset
        self.mult_factor = 2**pf.min_level
        self.min_left_edge = self._convert_coords(min_left)
        self.max_right_edge = self._convert_coords(max_right)
        self.min_left, self.max_right = min_left, max_right
        max_level = pf.max_level
        if max_level == -1: max_level = pf.base_pf.h.max_level
        self.max_level = min(max_level, pf.base_pf.h.max_level)
        self.final_level = self.max_level - self.min_level

        # Now we utilize the existing machinery for generating the appropriate
        # arrays of grids, etc etc.
        self.base_pf = pf.base_pf
        AMRHierarchy.__init__(self, pf, data_style)

        # Now a few cleanups
        self.pf.override["DomainRightEdge"] = self.max_right_edge
        self.pf.override["DomainLeftEdge"] = self.min_left_edge
        for u,v in self.base_pf.units.items():
            self.pf.override[u] = v / self.mult_factor
        self.pf.override['unitary'] = 1.0 / (self.pf.domain_right_edge -
                                             self.pf.domain_left_edge).max()

    def _count_grids(self):
        self.num_grids = 1 + sum( ( # 1 is the base grid
            len(self.base_pf.h.select_grids(level)) 
                for level in range(self.min_level+1, self.max_level)) )

    def _parse_hierarchy(self):
        # Here we need to set up the grid info, which for the Enzo hierarchy
        # is done like:
        # self.grid_dimensions.flat[:] = ei
        # self.grid_dimensions -= np.array(si, self.float_type)
        # self.grid_dimensions += 1
        # self.grid_left_edge.flat[:] = LE
        # self.grid_right_edge.flat[:] = RE
        # self.grid_particle_count.flat[:] = np
        # self.grids = np.array(self.grids, dtype='object')
        #
        # For now, we make the presupposition that all of our grids are
        # strictly nested and we are not doing any cuts.  However, we do
        # construct a root grid!
        root_level_grids = self.base_pf.h.select_grids(self.min_level)
        base_grid = ConstructedRootGrid(self.base_pf, self.pf, self,
                        self.min_level, self.min_left, self.max_right)
        self._fill_grid_arrays(base_grid, 0)
        grids = [base_grid]
        # We need to ensure we have the correct parentage relationships
        # However, we want the parent/child to be to the new proxy grids
        # so we need to map between the old ids and the new ids
        self.id_map = {}
        grid_id = 2 # id 0 is the base grid
        for level in range(self.min_level+1, self.max_level):
            for grid in self.base_pf.h.select_grids(level):
                # This next little bit will have to be changed if we ever move to
                # not-strictly-nested AMR hierarchies
                parent = self.id_map.get(grid.Parent.id, base_grid)
                grids.append(self.grid(grid_id, grid, self))
                parent.Children.append(grids[-1])
                grids[-1].Parent = parent
                self.id_map[grid.id] = grids[-1]
                # Now we fill in our arrays of values -- note that we
                # are filling in values from the base grids, not the newly
                # extracted grids.  We will perform bulk changes after we
                # finish.
                self._fill_grid_arrays(grid, grid_id-1)
                grid_id += 1

        self.grid_left_edge = self._convert_coords(self.grid_left_edge)
        self.grid_right_edge = self._convert_coords(self.grid_right_edge)
        self.grids = np.array(grids, dtype='object')

    def _fill_grid_arrays(self, grid, i):
        # This just fills in the grid arrays for a single grid --
        # note that we presuppose here that we are being handed a grid object
        # that has these defined; this means we are being handed the *base*
        # grid, not the newly extracted one
        self.grid_dimensions[i,:] = grid.ActiveDimensions
        self.grid_left_edge[i,:] = grid.LeftEdge
        self.grid_right_edge[i,:] = grid.RightEdge
        self.grid_particle_count[i] = grid.NumberOfParticles

    def _populate_grid_objects(self):
        for grid in self.grids:
            grid.Level = grid.base_grid.Level - self.pf.min_level
            grid._prepare_grid()
            grid._setup_dx()
            grid.start_index = None
        self.max_level -= self.pf.min_level
        print "New max level:", self.max_level

    def _convert_coords(self, val):
        return (val - self.left_edge_offset)*self.mult_factor

    def _detect_fields(self):
        self.field_list = self.base_pf.h.field_list[:]

    def _setup_unknown_fields(self):
        pass # Done in the base_h

    def _setup_derived_fields(self):
        self.derived_field_list = self.base_pf.h.derived_field_list[:]

    def _initialize_data_storage(self):
        self._data_file = None

    def _setup_classes(self):
        dd = self._get_data_reader_dict()
        AMRHierarchy._setup_classes(self, dd)
        self.object_types.sort()

class ExtractedParameterFile(StaticOutput):
    _hierarchy_class = ExtractedHierarchy
    data_style = "extracted"
    
    def __init__(self, base_pf, min_level, max_level = -1, offset = None):
        self.base_pf = base_pf
        self.min_level = min_level
        self.max_level = max_level
        self.offset = offset
        self.override = {}

    def __repr__(self):
        return "extracted_%s" % self.base_pf

    def __getattr__(self, name):
        # This won't get called if 'name' is found already
        # and we'd like it to raise AttributeError if it's not anywhere
        if name in ['h', 'hierarchy']:
            return StaticOutput._get_hierarchy(self)
        return getattr(self.base_pf, name)

    def __getitem__(self, key):
        if key not in self.override:
            return self.base_pf[key]
        return self.override[key]




"""
Simle integrators for the radiative transfer equation

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: KIPAC/SLAC/Stanford
Homepage: http://yt-project.org/
License:
  Copyright (C) 2008 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free, abs

cdef extern from "math.h":
    double exp(double x)
    float expf(float x)

@cython.boundscheck(False)
def Transfer3D(np.ndarray[np.float64_t, ndim=3] i_s,
               np.ndarray[np.float64_t, ndim=4] o_s,
               np.ndarray[np.float64_t, ndim=4] e,
               np.ndarray[np.float64_t, ndim=4] a,
               int imin, int imax, int jmin, int jmax,
               int kmin, int kmax, int istride, int jstride,
               np.float64_t dx):
    """
    This function accepts an incoming slab (*i_s*), a buffer
    for an outgoing set of values at every point in the grid (*o_s*),
    an emission array (*e*), an absorption array (*a*), and dimensions of
    the grid (*imin*, *imax*, *jmin*, *jmax*, *kmin*, *kmax*) as well
    as strides in the *i* and *j* directions, and a *dx* of the grid being
    integrated.
    """
    cdef int i, ii
    cdef int j, jj
    cdef int k, kk
    cdef int n, nn
    nn = o_s.shape[3] # This might be slow
    cdef np.float64_t *temp = <np.float64_t *>malloc(sizeof(np.float64_t) * nn)
    for i in range((imax-imin)*istride):
        ii = i + imin*istride
        for j in range((jmax-jmin)*jstride):
            jj = j + jmin*jstride
            # Not sure about the ordering of the loops here
            for n in range(nn):
                temp[n] = i_s[ii,jj,n] 
            for k in range(kmax-kmin):
                kk = k + kmin#*kstride, which doesn't make any sense
                for n in range(nn):
                    o_s[i,j,k,n] = temp[n] + dx*(e[i,j,k,n] - temp[n]*a[i,j,k,n])
                    temp[n] = o_s[i,j,k,n]
            for n in range(nn):
                i_s[ii,jj,n] = temp[n]
    free(temp)

@cython.boundscheck(False)
def TransferShells(np.ndarray[np.float64_t, ndim=3] i_s,
                   np.ndarray[np.float64_t, ndim=3] data,
                   np.ndarray[np.float64_t, ndim=2] shells):
    """
    This function accepts an incoming slab (*i_s*), a buffer of *data*,
    and a list of shells specified as [ (value, tolerance, r, g, b), ... ].
    """
    cdef int i, ii
    cdef int j, jj
    cdef int k, kk
    cdef int n, nn
    cdef np.float64_t dist
    ii = data.shape[0]
    jj = data.shape[1]
    kk = data.shape[2]
    nn = shells.shape[0]
    cdef float rgba[4]
    cdef float alpha
    for i in range(ii):
        for j in range(jj):
            # Not sure about the ordering of the loops here
            for k in range(kk):
                for n in range(nn):
                    dist = shells[n, 0] - data[i,j,k]
                    if dist < 0: dist *= -1.0
                    if dist < shells[n,1]:
                        dist = exp(-dist/8.0)
                        rgba[0] = shells[n,2]
                        rgba[1] = shells[n,3]
                        rgba[2] = shells[n,4]
                        rgba[3] = shells[n,5]
                        alpha = i_s[i,j,3]
                        dist *= dist # This might improve appearance
                        i_s[i,j,0] += (1.0 - alpha)*rgba[0]*dist*rgba[3]
                        i_s[i,j,1] += (1.0 - alpha)*rgba[1]*dist*rgba[3]
                        i_s[i,j,2] += (1.0 - alpha)*rgba[2]*dist*rgba[3]
                        i_s[i,j,3] += (1.0 - alpha)*rgba[3]*dist*rgba[3]
                        break

@cython.boundscheck(False)
def Transfer1D(float i_s,
               np.ndarray[np.float_t, ndim=1] o_s,
               np.ndarray[np.float_t, ndim=1] e,
               np.ndarray[np.float_t, ndim=1] a,
               np.ndarray[np.float_t, ndim=1] dx,
               int imin, int imax):
    cdef int i
    for i in range(imin, imax):
        o_s[i] = i_s + dx[i]*(e[i] - i_s*a[i])
        i_s = o_s[i]
    return i_s

@cython.wraparound(False)
@cython.boundscheck(False)
def VoxelTraversal(np.ndarray[np.int_t, ndim=3] grid_mask,
                   np.ndarray[np.float64_t, ndim=3] grid_t,
                   np.ndarray[np.float64_t, ndim=3] grid_dt,
                   np.ndarray[np.float64_t, ndim=1] left_edge,
                   np.ndarray[np.float64_t, ndim=1] right_edge,
                   np.ndarray[np.float64_t, ndim=1] dx,
                   np.ndarray[np.float64_t, ndim=1] u,
                   np.ndarray[np.float64_t, ndim=1] v):
    # We're roughly following Amanatides & Woo
    # Find the first place the ray hits the grid on its path
    # Do left edge then right edge in each dim
    cdef int i, x, y
    cdef np.float64_t tl, tr, intersect_t, enter_t, exit_t, dt_tolerance
    cdef np.float64_t iv_dir[3], tdelta[3], tmax[3], intersect[3]
    cdef np.int64_t step[3], cur_ind[3]
    intersect_t = 1
    dt_tolerance = 1e-6
    # recall p = v * t + u
    #  where p is position, v is our vector, u is the start point
    for i in range(3):
        # As long as we're iterating, set some other stuff, too
        if(v[i] < 0):
            step[i] = -1
        elif (v[i] == 0):
            step[i] = 1
            tmax[i] = 1e60
            iv_dir[i] = 1e60
            tdelta[i] = 1e-60
            continue
        else:
            step[i] = 1
        x = (i+1)%3
        y = (i+2)%3
        iv_dir[i] = 1.0/v[i]
        tl = (left_edge[i] - u[i])*iv_dir[i]
        tr = (right_edge[i] - u[i])*iv_dir[i]
        if (left_edge[x] <= (u[x] + tl*v[x]) <= right_edge[x]) and \
           (left_edge[y] <= (u[y] + tl*v[y]) <= right_edge[y]) and \
           (0.0 <= tl < intersect_t):
            intersect_t = tl
        if (left_edge[x] <= (u[x] + tr*v[x]) <= right_edge[x]) and \
           (left_edge[y] <= (u[y] + tr*v[y]) <= right_edge[y]) and \
           (0.0 <= tr < intersect_t):
            intersect_t = tr
    # if fully enclosed
    if (left_edge[0] <= u[0] <= right_edge[0]) and \
       (left_edge[1] <= u[1] <= right_edge[1]) and \
       (left_edge[2] <= u[2] <= right_edge[2]):
        intersect_t = 0.0
    if not (0 <= intersect_t <= 1): return
    # Now get the indices of the intersection
    for i in range(3):
        intersect[i] = u[i] + intersect_t * v[i]
        cur_ind[i] = np.floor((intersect[i] + 1e-8*dx[i] - left_edge[i])/dx[i])
        tmax[i] = (((cur_ind[i]+step[i])*dx[i])+left_edge[i]-u[i])*iv_dir[i]
        if cur_ind[i] == grid_mask.shape[i] and step[i] < 0:
            cur_ind[i] = grid_mask.shape[i] - 1
        if step[i] > 0: tmax[i] = (((cur_ind[i]+1)*dx[i])+left_edge[i]-u[i])*iv_dir[i]
        if step[i] < 0: tmax[i] = (((cur_ind[i]+0)*dx[i])+left_edge[i]-u[i])*iv_dir[i]
        tdelta[i] = (dx[i]*iv_dir[i])
        if tdelta[i] < 0: tdelta[i] *= -1
    # The variable intersect contains the point we first pierce the grid
    enter_t = intersect_t
    while 1:
        if (not (0 <= cur_ind[0] < grid_mask.shape[0])) or \
           (not (0 <= cur_ind[1] < grid_mask.shape[1])) or \
           (not (0 <= cur_ind[2] < grid_mask.shape[2])):
            break
        # Note that we are calculating t on the fly, but we get *negative* t
        # values from what they should be.
        # If we've reached t = 1, we are done.
        grid_mask[cur_ind[0], cur_ind[1], cur_ind[2]] = 1
        if (tmax[0] > 1.0) and (tmax[1] > 1.0) and (tmax[2] > 1.0):
            grid_t[cur_ind[0], cur_ind[1], cur_ind[2]] = 1.0
            grid_dt[cur_ind[0], cur_ind[1], cur_ind[2]] = 1.0 - enter_t
            break
        if tmax[0] < tmax[1]:
            if tmax[0] < tmax[2]:
                grid_t[cur_ind[0], cur_ind[1], cur_ind[2]] = enter_t
                grid_dt[cur_ind[0], cur_ind[1], cur_ind[2]] = tmax[0] - enter_t
                enter_t = tmax[0]
                tmax[0] += tdelta[0]
                cur_ind[0] += step[0]
            else:
                grid_t[cur_ind[0], cur_ind[1], cur_ind[2]] = enter_t
                grid_dt[cur_ind[0], cur_ind[1], cur_ind[2]] = tmax[2] - enter_t
                enter_t = tmax[2]
                tmax[2] += tdelta[2]
                cur_ind[2] += step[2]
        else:
            if tmax[1] < tmax[2]:
                grid_t[cur_ind[0], cur_ind[1], cur_ind[2]] = enter_t
                grid_dt[cur_ind[0], cur_ind[1], cur_ind[2]] = tmax[1] - enter_t
                enter_t = tmax[1]
                tmax[1] += tdelta[1]
                cur_ind[1] += step[1]
            else:
                grid_t[cur_ind[0], cur_ind[1], cur_ind[2]] = enter_t
                grid_dt[cur_ind[0], cur_ind[1], cur_ind[2]] = tmax[2] - enter_t
                enter_t = tmax[2]
                tmax[2] += tdelta[2]
                cur_ind[2] += step[2]
    return

@cython.wraparound(False)
@cython.boundscheck(False)
def PlaneVoxelIntegration(np.ndarray[np.float64_t, ndim=1] left_edge,
                          np.ndarray[np.float64_t, ndim=1] right_edge,
                          np.ndarray[np.float64_t, ndim=1] dx,
                          np.ndarray[np.float64_t, ndim=2] ug,
                          np.ndarray[np.float64_t, ndim=1] v,
                          np.ndarray[np.float64_t, ndim=2] image,
                          np.ndarray[np.float64_t, ndim=3] data,
                          np.ndarray[np.float64_t, ndim=2] shells):
    # We're roughly following Amanatides & Woo on a ray-by-ray basis
    # Note that for now it's just shells, but this can and should be
    # generalized to transfer functions
    cdef int i, x, y, vi
    intersect_t = 1
    dt_tolerance = 1e-6
    cdef int nv = ug.shape[0]
    cdef int nshells = shells.shape[0]
    cdef np.ndarray[np.float64_t, ndim=1] u = np.empty((3,), dtype=np.float64)
    # Copy things into temporary location for passing between functions
    for vi in range(nv):
        for i in range(3): u[i] = ug[vi, i]
        integrate_ray(u, v, left_edge, right_edge, dx, 
                      nshells, vi, data, shells, image)

@cython.wraparound(False)
@cython.boundscheck(False)
def integrate_ray(np.ndarray[np.float64_t, ndim=1] u,
                  np.ndarray[np.float64_t, ndim=1] v,
                  np.ndarray[np.float64_t, ndim=1] left_edge,
                  np.ndarray[np.float64_t, ndim=1] right_edge,
                  np.ndarray[np.float64_t, ndim=1] dx,
                  int nshells, int ind,
                  np.ndarray[np.float64_t, ndim=3] data,
                  np.ndarray[np.float64_t, ndim=2] shells,
                  np.ndarray[np.float64_t, ndim=2] image):
    cdef int step[3], x, y, i, n
    cdef np.float64_t intersect_t = 1
    cdef np.float64_t dt_tolerance = 1e-6
    cdef np.float64_t tl, tr, enter_t, exit_t
    cdef np.int64_t cur_ind[3]
    cdef np.float64_t tdelta[3]
    cdef np.float64_t tmax[3]
    cdef np.float64_t intersect[3]
    cdef np.float64_t dt, dv
    cdef np.float64_t dist, alpha
    cdef np.float64_t one = 1.0
    cdef int dims[3]
    cdef np.float64_t rgba[4], temp_x, temp_y
    for i in range(3):
        # As long as we're iterating, set some other stuff, too
        dims[i] = data.shape[i]
        if(v[i] < 0): step[i] = -1
        else: step[i] = 1
        x = (i+1)%3
        y = (i+2)%3
        tl = (left_edge[i] - u[i])/v[i]
        tr = (right_edge[i] - u[i])/v[i]
        temp_x = (u[x] + tl*v[x])
        temp_y = (u[y] + tl*v[y])
        if (left_edge[x] <= temp_x) and (temp_x <= right_edge[x]) and \
           (left_edge[y] <= temp_y) and (temp_y <= right_edge[y]) and \
           (0.0 <= tl) and (tl < intersect_t):
            intersect_t = tl
        temp_x = (u[x] + tr*v[x])
        temp_y = (u[y] + tr*v[y])
        if (left_edge[x] <= temp_x) and (temp_x <= right_edge[x]) and \
           (left_edge[y] <= temp_y) and (temp_y <= right_edge[y]) and \
           (0.0 <= tr) and (tr < intersect_t):
            intersect_t = tr
    # if fully enclosed
    if (left_edge[0] <= u[0] <= right_edge[0]) and \
       (left_edge[1] <= u[1] <= right_edge[1]) and \
       (left_edge[2] <= u[2] <= right_edge[2]):
        intersect_t = 0.0
    if not (0 <= intersect_t <= 1):
        #print "Returning: intersect_t ==", intersect_t
        return
    # Now get the indices of the intersection
    for i in range(3): intersect[i] = u[i] + intersect_t * v[i]
    cdef int ncells = 0
    for i in range(3):
        cur_ind[i] = np.floor((intersect[i] + 1e-8*dx[i] - left_edge[i])/dx[i])
        tmax[i] = (((cur_ind[i]+step[i])*dx[i])+left_edge[i]-u[i])/v[i]
        if cur_ind[i] == dims[i] and step[i] < 0:
            cur_ind[i] = dims[i] - 1
        if step[i] > 0: tmax[i] = (((cur_ind[i]+1)*dx[i])+left_edge[i]-u[i])/v[i]
        if step[i] < 0: tmax[i] = (((cur_ind[i]+0)*dx[i])+left_edge[i]-u[i])/v[i]
        tdelta[i] = (dx[i]/v[i])
        if tdelta[i] < 0: tdelta[i] *= -1
    # The variable intersect contains the point we first pierce the grid
    enter_t = intersect_t
    if (not (0 <= cur_ind[0] < dims[0])) or \
       (not (0 <= cur_ind[1] < dims[1])) or \
       (not (0 <= cur_ind[2] < dims[2])):
        #print "Returning: cur_ind", cur_ind[0], cur_ind[1], cur_ind[2]
        #print "  dims:     ", dims[0], dims[1], dims[2]
        #print "  intersect:",  intersect[0], intersect[1], intersect[2]
        #print "  intersect:", intersect_t
        #print "  u        :", u[0], u[1], u[2]
        #
        return
    #print cur_ind[0], dims[0], cur_ind[1], dims[1], cur_ind[2], dims[2]
    dv = data[cur_ind[0], cur_ind[1], cur_ind[2]]
    #dt = 1e300
    while 1:
        if image[ind,3] >= 1.0: break
        if (not (0 <= cur_ind[0] < dims[0])) or \
           (not (0 <= cur_ind[1] < dims[1])) or \
           (not (0 <= cur_ind[2] < dims[2])):
            break
        # Do our transfer here
        for n in range(nshells):
            dist = shells[n, 0] - dv
            if dist < shells[n,1]:
                dist = exp(-dist/8.0)
                alpha = (1.0 - shells[n,5])*shells[n,5]#*dt
                image[ind,0] += alpha*shells[n,2]*dist
                image[ind,1] += alpha*shells[n,3]*dist
                image[ind,2] += alpha*shells[n,4]*dist
                image[ind,3] += alpha*shells[n,5]*dist
                #image[ind,i] += rgba[i]*dist*rgba[3]/dt
                #print rgba[i], image[ind,i], dist, dt
                break
        if (tmax[0] > 1.0) and (tmax[1] > 1.0) and (tmax[2] > 1.0):
            dt = 1.0 - enter_t
            break
        if tmax[0] < tmax[1]:
            if tmax[0] < tmax[2]:
                dt = tmax[0] - enter_t
                enter_t = tmax[0]
                tmax[0] += tdelta[0]
                cur_ind[0] += step[0]
            else:
                dt = tmax[2] - enter_t
                enter_t = tmax[2]
                tmax[2] += tdelta[2]
                cur_ind[2] += step[2]
        else:
            if tmax[1] < tmax[2]:
                dt = tmax[1] - enter_t
                enter_t = tmax[1]
                tmax[1] += tdelta[1]
                cur_ind[1] += step[1]
            else:
                dt = tmax[2] - enter_t
                enter_t = tmax[2]
                tmax[2] += tdelta[2]
                cur_ind[2] += step[2]
        dv = data[cur_ind[0], cur_ind[1], cur_ind[2]]



"""
Skeleton objects that represent a few fundamental yt data types.

Author: Matthew Turk <matthewturk@gmail.com>
Affiliation: Columbia University
Homepage: http://yt-project.org/
License:
  Copyright (C) 2011 Matthew Turk.  All Rights Reserved.

  This file is part of yt.

  yt is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import abc
import json
import urllib2
from tempfile import TemporaryFile
from yt.config import ytcfg
from yt.funcs import *
from yt.utilities.exceptions import *

from .poster.streaminghttp import register_openers
from .poster.encode import multipart_encode
register_openers()

class UploaderBar(object):
    pbar = None
    def __init__(self, my_name = ""):
        self.my_name = my_name

    def __call__(self, name, prog, total):
        if self.pbar is None:
            self.pbar = get_pbar("Uploading %s " % self.my_name, total)
        self.pbar.update(prog)
        if prog == total:
            self.pbar.finish()

class ContainerClass(object):
    pass

class MinimalRepresentation(object):
    __metaclass__ = abc.ABCMeta

    def _update_attrs(self, obj, attr_list):
        for attr in attr_list:
            setattr(self, attr, getattr(obj, attr, None))
        if hasattr(obj, "pf"):
            self.output_hash = obj.pf._hash()
            self._pf_mrep = obj.pf._mrep

    def __init__(self, obj):
        self._update_attrs(obj, self._attr_list)

    @abc.abstractmethod
    def _generate_post(self):
        pass

    @abc.abstractproperty
    def _attr_list(self):
        pass

    def _return_filtered_object(self, attrs):
        new_attrs = tuple(attr for attr in self._attr_list
                          if attr not in attrs)
        new_class = type('Filtered%s' % self.__class__.__name__,
                         (FilteredRepresentation,),
                         {'_attr_list': new_attrs})
        return new_class(self)

    @property
    def _attrs(self):
        return dict( ((attr, getattr(self, attr)) for attr in self._attr_list) )

    @classmethod
    def _from_metadata(cls, metadata):
        cc = ContainerClass()
        for a, v in metadata.values():
            setattr(cc, a, v)
        return cls(cc)

    def upload(self):
        api_key = ytcfg.get("yt","hub_api_key")
        url = ytcfg.get("yt","hub_url")
        if api_key == '': raise YTHubRegisterError
        metadata, (final_name, chunks) = self._generate_post()
        if hasattr(self, "_pf_mrep"):
            self._pf_mrep.upload()
        for i in metadata:
            if isinstance(metadata[i], np.ndarray):
                metadata[i] = metadata[i].tolist()
            elif hasattr(metadata[i], 'dtype'):
                metadata[i] = np.asscalar(metadata[i])
        metadata['obj_type'] = self.type
        if len(chunks) == 0:
            chunk_info = {'chunks': []}
        else:
            chunk_info = {'final_name' : final_name, 'chunks': []}
            for cn, cv in chunks:
                chunk_info['chunks'].append((cn, cv.size * cv.itemsize))
        metadata = json.dumps(metadata)
        chunk_info = json.dumps(chunk_info)
        datagen, headers = multipart_encode({'metadata' : metadata,
                                             'chunk_info' : chunk_info,
                                             'api_key' : api_key})
        request = urllib2.Request(url, datagen, headers)
        # Actually do the request, and get the response
        try:
            rv = urllib2.urlopen(request).read()
        except urllib2.HTTPError as ex:
            if ex.code == 401:
                mylog.error("You must create an API key before uploading.")
                mylog.error("https://data.yt-project.org/getting_started.html")
                return
            else:
                raise ex
        uploader_info = json.loads(rv)
        new_url = url + "/handler/%s" % uploader_info['handler_uuid']
        for i, (cn, cv) in enumerate(chunks):
            remaining = cv.size * cv.itemsize
            f = TemporaryFile()
            np.save(f, cv)
            f.seek(0)
            pbar = UploaderBar("%s, % 2i/% 2i" % (self.type, i+1, len(chunks)))
            datagen, headers = multipart_encode({'chunk_data' : f}, cb = pbar)
            request = urllib2.Request(new_url, datagen, headers)
            rv = urllib2.urlopen(request).read()

        datagen, headers = multipart_encode({'status' : 'FINAL'})
        request = urllib2.Request(new_url, datagen, headers)
        rv = json.loads(urllib2.urlopen(request).read())
        mylog.info("Upload succeeded!  View here: %s", rv['url'])
        return rv

class FilteredRepresentation(MinimalRepresentation):
    def _generate_post(self):
        raise RuntimeError

class MinimalStaticOutput(MinimalRepresentation):
    _attr_list = ("dimensionality", "refine_by", "domain_dimensions",
                  "current_time", "domain_left_edge", "domain_right_edge",
                  "unique_identifier", "current_redshift", "output_hash",
                  "cosmological_simulation", "omega_matter", "omega_lambda",
                  "hubble_constant", "name")
    type = 'simulation_output'

    def __init__(self, obj):
        super(MinimalStaticOutput, self).__init__(obj)
        self.output_hash = obj._hash()
        self.name = str(obj)

    def _generate_post(self):
        metadata = self._attrs
        chunks = []
        return (metadata, (None, chunks))

class MinimalMappableData(MinimalRepresentation):

    _attr_list = ("field_data", "field", "weight_field", "axis", "output_hash",
                  "vm_type")

    def _generate_post(self):
        nobj = self._return_filtered_object(("field_data",))
        metadata = nobj._attrs
        chunks = [(arr, self.field_data[arr]) for arr in self.field_data]
        return (metadata, ('field_data', chunks))

class MinimalProjectionData(MinimalMappableData):
    type = 'proj'
    vm_type = "Projection"

class MinimalSliceData(MinimalMappableData):
    type = 'slice'
    vm_type = "Slice"
    weight_field = "None"

class MinimalImageCollectionData(MinimalRepresentation):
    type = "image_collection"
    _attr_list = ("name", "output_hash", "images", "image_metadata")

    def _generate_post(self):
        nobj = self._return_filtered_object(("images",))
        metadata = nobj._attrs
        chunks = [(fn, d) for fn, d in self.images]
        return (metadata, ('images', chunks))

_hub_categories = ("News", "Documents", "Simulation Management",
                   "Data Management", "Analysis and Visualization",
                   "Paper Repositories", "Astrophysical Utilities",
                   "yt Scripts")

class MinimalProjectDescription(MinimalRepresentation):
    type = "project"
    _attr_list = ("title", "url", "description", "category", "image_url")

    def __init__(self, title, url, description,
                 category, image_url = ""):
        assert(category in _hub_categories)
        self.title = title
        self.url = url
        self.description = description
        self.category = category
        self.image_url = image_url

    def _generate_post(self):
        metadata = self._attrs
        chunks = []
        return (metadata, ("chunks", []))

class MinimalNotebook(MinimalRepresentation):
    type = "notebook"
    _attr_list = ("title",)

    def __init__(self, filename, title = None):
        # First we read in the data
        if not os.path.isfile(filename):
            raise IOError(filename)
        self.data = open(filename).read()
        if title is None:
            title = json.loads(self.data)['metadata']['name']
        self.title = title
        self.data = np.fromstring(self.data, dtype='c')

    def _generate_post(self):
        metadata = self._attrs
        chunks = [ ("notebook", self.data) ]
        return (metadata, ("chunks", chunks))
